from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Iterable, Iterator, Optional

import torch

from ..data import DataLoaderBase


class Evaluator(metaclass=ABCMeta):
    """
    Base class for in-loop evaluators.

    .. seealso::
        This can be used with an :class:`~olmo_core.train.callbacks.EvaluatorCallback` to run an
        evaluator within the training loop.

    :param name: A name to assign to the evaluator.
    :param batches: Generates batches for the evaluator. These should at least include the
        "input_ids" field, but can contain any other arbitrary fields as well.
    :param device: The device to compute/reduce metrics on.
    """

    def __init__(
        self,
        *,
        name: str,
        batches: Iterable[Dict[str, Any]],
        device: Optional[torch.device] = None,
    ):
        self.name = name
        self.batches = batches
        self.device = device

    def __iter__(self) -> Iterator[Dict[str, Any]]:
        """
        Iterator over the evaluator's batches.
        """
        if isinstance(self.batches, DataLoaderBase):
            self.batches.reshuffle(in_memory=True)
        for batch in self.batches:
            yield batch
        if isinstance(self.batches, DataLoaderBase):
            self.batches.reset()

    @property
    def total_batches(self) -> Optional[int]:
        """
        Get the total number of batches in an eval loop if it's known ahead of time.
        """
        try:
            return len(self.batches)  # type: ignore
        except TypeError:
            return None

    @abstractmethod
    def update_metrics(
        self, batch: Dict[str, Any], ce_loss: Optional[torch.Tensor], logits: Optional[torch.Tensor]
    ) -> None:
        """
        Update metrics with from the ``batch`` just processed and the corresponding ``logits``.

        :param batch: A batch generated from :data:`batches`.
        :param ce_loss: The cross-entropy loss per token (un-reduced) of the batch. This will
            have shape ``(batch_size, (seq_len - 1))``.
        :param logits: The logits generated from the forward pass of the model.
        """
        raise NotImplementedError

    @abstractmethod
    def compute_metrics(self) -> Dict[str, torch.Tensor]:
        """
        Compute the final value of the metrics for the current evaluation loop.
        The metrics returned should already be reduced, if needed.
        """
        raise NotImplementedError

    @abstractmethod
    def reset_metrics(self) -> None:
        """
        Reset metrics. Should be called after :meth:`compute_metrics()`.
        """
        raise NotImplementedError



from .adam import AdamConfig
from .adamw import AdamWConfig, SkipStepAdamW, SkipStepAdamWConfig
from .config import INITIAL_LR_FIELD, LR_FIELD, OptimConfig, OptimGroupOverride
from .lion import Lion, LionConfig, SkipStepLion, SkipStepLionConfig
from .scheduler import (
    WSD,
    ConstantScheduler,
    ConstantWithWarmup,
    CosWithWarmup,
    HalfCosWithWarmup,
    InvSqrtWithWarmup,
    LinearWithWarmup,
    Scheduler,
    SchedulerUnits,
    SequentialScheduler,
)
from .skip_step_optimizer import SkipStepOptimizer

__all__ = [
    "OptimConfig",
    "OptimGroupOverride",
    "SkipStepOptimizer",
    "AdamWConfig",
    "SkipStepAdamWConfig",
    "SkipStepAdamW",
    "AdamConfig",
    "LionConfig",
    "Lion",
    "SkipStepLionConfig",
    "SkipStepLion",
    "Scheduler",
    "SchedulerUnits",
    "ConstantScheduler",
    "ConstantWithWarmup",
    "CosWithWarmup",
    "HalfCosWithWarmup",
    "InvSqrtWithWarmup",
    "LinearWithWarmup",
    "SequentialScheduler",
    "WSD",
    "LR_FIELD",
    "INITIAL_LR_FIELD",
]



import json
import logging
import os
import platform
import subprocess
import time
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional

from olmo_core.distributed.utils import get_rank
from olmo_core.exceptions import OLMoEnvironmentError

from ..common import TrainingProgress
from .callback import Callback
from .comet import CometCallback
from .wandb import WandBCallback

if TYPE_CHECKING:
    from beaker import Beaker

log = logging.getLogger(__name__)


BEAKER_EXPERIMENT_ID_ENV_VAR = "BEAKER_EXPERIMENT_ID"
BEAKER_RESULT_DIR = "/results"


@dataclass
class BeakerCallback(Callback):
    """
    Adds metadata to the Beaker experiment description when running as a Beaker batch job.
    """

    priority: ClassVar[int] = min(CometCallback.priority - 1, WandBCallback.priority - 1)
    experiment_id: Optional[str] = None
    update_interval: Optional[int] = None
    description: Optional[str] = None
    enabled: Optional[bool] = None
    config: Optional[Dict[str, Any]] = None
    """
    A JSON-serializable config to save to the results dataset as ``config.json``.
    """
    result_dir: str = BEAKER_RESULT_DIR
    """
    The directory of the Beaker results dataset where the config and other data will be saved.
    """

    _client = None
    _url = None
    _last_update: Optional[float] = None

    @property
    def client(self) -> "Beaker":
        return self._client  # type: ignore

    @client.setter
    def client(self, client: "Beaker"):
        self._client = client

    def post_attach(self):
        if self.enabled is None and BEAKER_EXPERIMENT_ID_ENV_VAR in os.environ:
            self.enabled = True

    def pre_train(self):
        if self.enabled and get_rank() == 0:
            if self.experiment_id is None:
                if BEAKER_EXPERIMENT_ID_ENV_VAR not in os.environ:
                    raise OLMoEnvironmentError(f"missing env var '{BEAKER_EXPERIMENT_ID_ENV_VAR}'")
                else:
                    self.experiment_id = os.environ[BEAKER_EXPERIMENT_ID_ENV_VAR]

            from beaker import Beaker

            self.client = Beaker.from_env()
            log.info(
                f"Running in Beaker experiment {self.client.experiment.url(self.experiment_id)}"
            )

            # Ensure result dataset directory exists.
            result_dir = Path(self.result_dir) / "olmo-core"
            result_dir.mkdir(parents=True, exist_ok=True)

            # Save config to result dir.
            if self.config is not None:
                config_path = result_dir / "config.json"
                with config_path.open("w") as config_file:
                    log.info(f"Saving config to '{config_path}'")
                    json.dump(self.config, config_file)

            # Try saving Python requirements.
            requirements_path = result_dir / "requirements.txt"
            try:
                with requirements_path.open("w") as requirements_file:
                    requirements_file.write(f"# python={platform.python_version()}\n")
                with requirements_path.open("a") as requirements_file:
                    subprocess.call(
                        ["pip", "freeze"],
                        stdout=requirements_file,
                        stderr=subprocess.DEVNULL,
                        timeout=10,
                    )
            except Exception as e:
                log.exception(f"Error saving Python packages: {e}")

            # Try to get W&B/Comet URL of experiment.
            for callback in self.trainer.callbacks.values():
                if isinstance(callback, WandBCallback) and callback.enabled:
                    if (url := callback.run.get_url()) is not None:
                        self._url = url
                    break
                elif isinstance(callback, CometCallback) and callback.enabled:
                    if (url := callback.exp.url) is not None:
                        self._url = url
                    break

            self._update()

    def post_step(self):
        update_interval = self.update_interval or self.trainer.metrics_collect_interval
        if self.enabled and get_rank() == 0 and self.step % update_interval == 0:
            # Make sure we don't update too frequently.
            if self._last_update is None or (time.monotonic() - self._last_update) > 10:
                self._update()

    def post_train(self):
        if self.enabled and get_rank() == 0:
            self._update()

    def _update(self):
        self.trainer.run_bookkeeping_op(
            self._set_description,
            self.trainer.training_progress,
            op_name="beaker_set_description",
            allow_multiple=False,
            distributed=False,
        )
        self._last_update = time.monotonic()

    def _set_description(self, progress: TrainingProgress):
        from beaker import BeakerError, HTTPError
        from requests.exceptions import RequestException

        assert self.experiment_id is not None

        description = f"[{progress}] "

        if self.description is not None:
            description = f"{description}{self.description}\n"

        if self._url is not None:
            description = f"{description}{self._url} "

        try:
            self.client.experiment.set_description(self.experiment_id, description.strip())
        except (RequestException, BeakerError, HTTPError) as e:
            log.warning(f"Failed to update Beaker experiment description: {e}")



from collections import defaultdict
from collections.abc import MutableMapping
from typing import Dict, Optional

import torch

from olmo_core.utils import move_to_device


class BufferCache(MutableMapping[str, torch.Tensor]):
    """
    Cache for buffers such as attention biases that would normally be registered as module buffers.

    We avoid using buffers because we've run into various issues doing so with FSDP.
    In general it appears the way FSDP handles buffers is not well-defined.
    It doesn't shard them but apparently it does synchronize them across processes, which we want to avoid
    since (A) it isn't necessary, and (B) we sometimes have `-inf` in these biases which might get turned into
    NaNs when they're synchronized due to casting or some other issue.

    :param namespace: Optional namespace for the cache. This allows you to have a separate sub-cache
        in a shared :class:`BufferCache` to avoid key collisions. See how this is used in the
        :class:`olmo_core.nn.rope.RotaryEmbeddingBase` class for an example.
    """

    def __init__(self, namespace: str = ""):
        self._data: Dict[str, Dict[str, torch.Tensor]] = defaultdict(dict)
        self._namespace = namespace

    def __getitem__(self, key: str) -> torch.Tensor:
        return self._data[self._namespace][key]

    def __setitem__(self, key: str, value: torch.Tensor) -> None:
        self._data[self._namespace][key] = value

    def __delitem__(self, key: str) -> None:
        del self._data[self._namespace][key]

    def __iter__(self):
        yield from self._data[self._namespace].keys()

    def __len__(self) -> int:
        return len(self._data[self._namespace])

    def get_for_device(self, key: str, device: torch.device) -> Optional[torch.Tensor]:
        if (tensor := self.get(key)) is not None:
            if tensor.device != device:
                tensor = move_to_device(tensor, device)
                self[key] = tensor
            return tensor
        else:
            return None

    def with_namespace(self, namespace: str) -> "BufferCache":
        """
        This creates a new :class:`BufferCache` object with a pointer to the same underlying data
        but with the given namespace.
        """
        out = BufferCache(namespace=namespace)
        out._data = self._data
        return out



from abc import ABCMeta, abstractmethod
from typing import Optional, Union

import torch
import torch.distributed as dist

from ..distributed.utils import all_reduce_value
from ..utils import get_default_device

__all__ = ["Metric", "MeanMetric"]


class Metric(metaclass=ABCMeta):
    """
    Base class for evaluation metrics.
    """

    def __init__(
        self,
        device: Optional[torch.device] = None,
        process_group: Optional[dist.ProcessGroup] = None,
    ):
        self.device = device if device is not None else get_default_device()
        self.process_group = process_group

    @abstractmethod
    def update(self, *args, **kwargs) -> None:
        """
        Update the metric.
        """
        raise NotImplementedError

    @abstractmethod
    def compute(self) -> torch.Tensor:
        """
        Compute the metric.
        """
        raise NotImplementedError

    @abstractmethod
    def reset(self) -> None:
        """
        Reset the metric.
        """
        raise NotImplementedError

    def as_tensor(self, value: Union[float, torch.Tensor]) -> torch.Tensor:
        if not isinstance(value, torch.Tensor):
            value = torch.tensor(value, dtype=torch.float32)
        return value.to(device=self.device, non_blocking=self.device.type != "cpu")


class MeanMetric(Metric):
    """
    Computes the mean over a stream of values.
    """

    def __init__(
        self,
        device: Optional[torch.device] = None,
        process_group: Optional[dist.ProcessGroup] = None,
    ):
        super().__init__(device=device, process_group=process_group)
        self.weighted_sum = torch.tensor(0.0, device=self.device)
        self.weight = torch.tensor(0.0, device=self.device)

    def update(
        self, value: Union[float, torch.Tensor], weight: Union[float, torch.Tensor] = 1.0
    ) -> None:
        """
        :param value: The latest value to update the metric with. Could be a tensor of values.
        :param weight: The corresponding weight(s) for the value. Should be the same shape as ``value``.
        """
        value = self.as_tensor(value)
        weight = torch.broadcast_to(self.as_tensor(weight), value.shape)
        if value.numel() == 0:
            return
        self.weighted_sum += (value * weight).sum()
        self.weight += weight.sum()

    def compute(self) -> torch.Tensor:
        """
        Computes the mean over the values and weights given.
        """
        weighted_sum = all_reduce_value(
            self.weighted_sum, device=self.device, group=self.process_group
        )
        weight = all_reduce_value(self.weight, device=self.device, group=self.process_group)
        return weighted_sum / weight

    def reset(self) -> None:
        self.weighted_sum.zero_()
        self.weight.zero_()



import io
import logging
import os
import pickle
import re
import shutil
import time
from collections import deque
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Generator, Optional, Tuple, Type, Union

try:
    from functools import cache
except ImportError:
    from functools import lru_cache as cache

import requests
import torch
from cached_path import cached_path
from cached_path.schemes import S3Client, SchemeClient, add_scheme_client
from rich.progress import track

from .aliases import PathOrStr
from .exceptions import (
    OLMoEnvironmentError,
    OLMoInvalidRangeRequestError,
    OLMoNetworkError,
    OLMoUploadError,
)

log = logging.getLogger(__name__)

############################################
## Unified API for local and remote files ##
############################################


def normalize_path(path: PathOrStr) -> str:
    """
    Normalize a path/URL.

    :param path: The path/URL to normalize.
    """
    return str(path).rstrip("/").replace("file://", "")


def join_path(path1: PathOrStr, path2: PathOrStr) -> PathOrStr:
    """
    Join two paths.

    :param path1: The first path.
    :param path2: The second path.

    :returns: The joined result.
    """
    if is_url(path1):
        return f"{normalize_path(path1)}/{normalize_path(path2)}"
    else:
        return Path(path1) / path2


def resource_path(folder: PathOrStr, fname: str, local_cache: Optional[PathOrStr] = None) -> Path:
    """
    Returns an actual path for local or remote file, potentially downloading it if a copy doesn't
    exist locally yet.
    """
    folder = normalize_path(folder)
    if local_cache is not None and (local_path := Path(local_cache) / fname).is_file():
        log.info(f"Found local cache of {fname} at {local_path}")
        return local_path
    else:
        return cached_path(f"{folder}/{fname}", quiet=True)


def is_url(path: PathOrStr) -> bool:
    """
    Check if a path is a URL.

    :param path: Path-like object to check.
    """
    path = normalize_path(path)
    return re.match(r"[a-z0-9]+://.*", str(path)) is not None


def get_file_size(path: PathOrStr) -> int:
    """
    Get the size of a local or remote file in bytes.

    :param path: Path/URL to the file.
    """
    path = normalize_path(path)

    if is_url(path):
        from urllib.parse import urlparse

        parsed = urlparse(str(path))
        if parsed.scheme == "gs":
            return _gcs_file_size(parsed.netloc, parsed.path.strip("/"))
        elif parsed.scheme in ("s3", "r2", "weka"):
            return _s3_file_size(parsed.scheme, parsed.netloc, parsed.path.strip("/"))
        elif parsed.scheme in ("http", "https"):
            return _http_file_size(str(path))
        elif parsed.scheme == "file":
            return get_file_size(str(path).replace("file://", "", 1))
        else:
            raise NotImplementedError(f"file size not implemented for '{parsed.scheme}' files")
    else:
        return os.stat(path).st_size


def get_bytes_range(path: PathOrStr, bytes_start: int, num_bytes: int) -> bytes:
    """
    Get a range of bytes from a local or remote file.

    :param source: Path/URL to the file.
    :param bytes_start: Byte offset to start at.
    :param num_bytes: Number of bytes to get.
    """
    path = normalize_path(path)

    if is_url(path):
        from urllib.parse import urlparse

        parsed = urlparse(str(path))
        if parsed.scheme == "gs":
            return _gcs_get_bytes_range(
                parsed.netloc, parsed.path.strip("/"), bytes_start, num_bytes
            )
        elif parsed.scheme in ("s3", "r2", "weka"):
            return _s3_get_bytes_range(
                parsed.scheme, parsed.netloc, parsed.path.strip("/"), bytes_start, num_bytes
            )
        elif parsed.scheme in ("http", "https"):
            return _http_get_bytes_range(str(path), bytes_start, num_bytes)
        elif parsed.scheme == "file":
            return get_bytes_range(str(path).replace("file://", "", 1), bytes_start, num_bytes)
        else:
            raise NotImplementedError(f"file size not implemented for '{parsed.scheme}' files")
    else:
        with open(path, "rb") as f:
            f.seek(bytes_start)
            return f.read(num_bytes)


def upload(source: PathOrStr, target: str, save_overwrite: bool = False, quiet: bool = False):
    """
    Upload source file to a target location on GCS or S3.

    :param source: Path to the file to upload.
    :param target: Target URL to upload to.
    :param save_overwrite: Overwrite any existing file.
    """
    from urllib.parse import urlparse

    source = Path(normalize_path(source))
    assert source.is_file()
    num_bytes = get_file_size(source)
    if not quiet:
        log.info(f"Uploading {_format_bytes(num_bytes)} from '{source}' to '{target}'...")
    parsed = urlparse(target)

    if parsed.scheme == "gs":
        _gcs_upload(source, parsed.netloc, parsed.path.strip("/"), save_overwrite=save_overwrite)
    elif parsed.scheme in ("s3", "r2", "weka"):
        _s3_upload(
            source,
            parsed.scheme,
            parsed.netloc,
            parsed.path.strip("/"),
            save_overwrite=save_overwrite,
        )
    else:
        raise NotImplementedError(f"Upload not implemented for '{parsed.scheme}' scheme")

    if not quiet:
        log.info(f"Uploaded {_format_bytes(num_bytes)} to '{target}'")


def copy_file(
    source: PathOrStr, target: PathOrStr, save_overwrite: bool = False, quiet: bool = False
):
    """
    Copy a file from ``source`` to ``target``.

    :param source: The path/URL to the source file.
    :param target: The path/URL to the target location.
    :param save_overwrite: Overwrite any existing file.

    :raises FileNotFoundError: If the ``source`` file doesn't exist.
    :raises FileExistsError: If the ``target`` already exists and ``save_overwrite=False``.
    """
    source = normalize_path(source)
    target = normalize_path(target)
    local_source = cached_path(source, quiet=True)
    if is_url(target):
        upload(local_source, target, save_overwrite=save_overwrite, quiet=quiet)
    else:
        target = Path(target)
        if file_exists(target):
            if not save_overwrite:
                raise FileExistsError(target)
            else:
                target.unlink(missing_ok=True)
        else:
            target.parent.mkdir(exist_ok=True, parents=True)
        try:
            os.link(local_source, target, follow_symlinks=True)
        except OSError:
            # Can't hard link across devices, so copy instead.
            shutil.copyfile(local_source, target, follow_symlinks=True)


def copy_dir(
    source: PathOrStr,
    target: PathOrStr,
    save_overwrite: bool = False,
    num_threads: Optional[int] = None,
    quiet: bool = False,
):
    """
    Copy a directory from ``source`` to ``target``.

    :param source: The path/URL to the source file.
    :param target: The path/URL to the target location.
    :param save_overwrite: Overwrite any existing files.
    :param num_threads: The number of threads to use.

    :raises FileNotFoundError: If the ``source`` dir doesn't exist.
    :raises FileExistsError: If any source files already exist in the ``target`` and ``save_overwrite=False``.
    """
    source = normalize_path(source)
    target = normalize_path(target)

    if num_threads is None:
        from .utils import get_default_thread_count

        num_threads = get_default_thread_count()

    with ThreadPoolExecutor(max_workers=num_threads) as executor:
        if not quiet:
            log.info(f"Collecting source files from '{source}' to copy to '{target}'...")

        futures = []
        for source_path in list_directory(source, recurse=True, include_dirs=False):
            assert source_path.startswith(source)
            relative_source_path = source_path.replace(source, "", 1).lstrip("/")
            target_path = join_path(target, relative_source_path)
            futures.append(
                executor.submit(
                    copy_file, source_path, target_path, save_overwrite=save_overwrite, quiet=True
                )
            )

        if not quiet:
            log.info(f"Collected {len(futures)} source files to copy")

        deque(
            track(
                (f.result() for f in as_completed(futures)),
                description="Copying source files...",
                disable=quiet,
                total=len(futures),
            ),
            maxlen=0,
        )


def dir_is_empty(dir: PathOrStr) -> bool:
    """
    Check if a local or remote directory is empty.
    This also returns true if the directory does not exist.

    :param dir: Path/URL to the directory.
    """
    try:
        next(list_directory(dir))
        return False
    except (StopIteration, FileNotFoundError):
        return True


def file_exists(path: PathOrStr) -> bool:
    """
    Check if a local or remote file exists.

    :param path: Path/URL to a file.
    """
    path = normalize_path(path)

    if is_url(path):
        from urllib.parse import urlparse

        parsed = urlparse(str(path))
        if parsed.scheme == "gs":
            try:
                _gcs_file_size(parsed.netloc, parsed.path.strip("/"))
            except FileNotFoundError:
                return False
            else:
                return True
        elif parsed.scheme in ("s3", "r2", "weka"):
            try:
                _s3_file_size(parsed.scheme, parsed.netloc, parsed.path.strip("/"))
            except FileNotFoundError:
                return False
            else:
                return True
        elif parsed.scheme in ("http", "https"):
            return _http_file_exists(str(path))
        else:
            raise NotImplementedError(f"file_exists not implemented for '{parsed.scheme}' files")
    else:
        return Path(path).exists()


def remove_file(path: PathOrStr):
    """
    Remove a local or remote file.

    :param path: The path or URL to the file.

    :raises FileNotFoundError: If the file doesn't exist.
    """
    path = normalize_path(path)

    if is_url(path):
        from urllib.parse import urlparse

        parsed = urlparse(str(path))
        if parsed.scheme == "gs":
            return _gcs_remove_file(parsed.netloc, parsed.path.strip("/"))
        elif parsed.scheme in ("s3", "r2", "weka"):
            return _s3_remove_file(parsed.scheme, parsed.netloc, parsed.path.strip("/"))
        else:
            raise NotImplementedError(f"remove_file not implemented for '{parsed.scheme}' files")
    else:
        Path(path).unlink()


def clear_directory(dir: PathOrStr, force: bool = False):
    """
    Clear out the contents of a local or remote directory.

    .. warning::
        This function is potentially very destructive!

        By default, for safety, this raise a :class:`ValueError` if you attempt to clear a remote
        directory too close to the root of the bucket. Set ``force=True`` to override.

    :param dir: Path/URL to the directory.
    :param force: See note about safety.
    """
    dir = normalize_path(dir)

    if is_url(dir):
        from urllib.parse import urlparse

        parsed = urlparse(str(dir))
        if parsed.scheme in ("s3", "r2", "weka", "gs"):
            prefix = parsed.path.strip("/")
            # For safety (so people don't accidentally delete a whole bunch of important data),
            # ensure prefix is at least 2 folders deep.
            if not force and prefix.count("/") < 2:
                raise ValueError(
                    "For safety, clearing a remote directory this close to the root of a bucket is "
                    "not allowed by default. To override this behavior set ``force=True``."
                )

        if parsed.scheme == "gs":
            prefix = parsed.path.strip("/")
            return _gcs_clear_directory(parsed.netloc, prefix)
        elif parsed.scheme in ("s3", "r2", "weka"):
            prefix = parsed.path.strip("/")
            return _s3_clear_directory(parsed.scheme, parsed.netloc, prefix)
        else:
            raise NotImplementedError(
                f"clear_directory not implemented for '{parsed.scheme}' folders"
            )
    elif Path(dir).is_dir():
        shutil.rmtree(dir, ignore_errors=True)


def list_directory(
    dir: PathOrStr, recurse: bool = False, include_files: bool = True, include_dirs: bool = True
) -> Generator[str, None, None]:
    """
    List the contents of a local or remote directory.
    If ``recurse=False``, only the immediate children of the directory are returned, otherwise
    every sub-folder is recursed into.

    :param dir: Path/URL to the directory.
    :param recurse: Whether to recurse into sub-folders.
    :param include_files: Include regular files in the results.
    :param include_dirs: Include directories in the results.

    :returns: A generator over paths in the directory. If the ``dir`` is a URL, the results will be
        full URLs. If the ``dir`` is a local path, the results will be of the form ``join_path(dir, p)``.

    :raises FileNotFoundError: If the ``source`` file doesn't exist.
    """
    dir = normalize_path(dir)

    if not is_url(dir):
        for p in Path(dir).iterdir():
            if (p.is_file() and include_files) or (p.is_dir() and include_dirs):
                yield str(p)
            if recurse and p.is_dir():
                yield from list_directory(
                    p, recurse=True, include_files=include_files, include_dirs=include_dirs
                )
    else:
        from urllib.parse import urlparse

        parsed = urlparse(dir)
        if parsed.scheme == "gs":
            yield from _gcs_list_directory(
                parsed.netloc,
                parsed.path.strip("/"),
                recurse=recurse,
                include_files=include_files,
                include_dirs=include_dirs,
            )
        elif parsed.scheme in ("s3", "r2", "weka"):
            yield from _s3_list_directory(
                parsed.scheme,
                parsed.netloc,
                parsed.path.strip("/"),
                recurse=recurse,
                include_files=include_files,
                include_dirs=include_dirs,
            )
        else:
            raise NotImplementedError(
                f"list_directory size not implemented for '{parsed.scheme}' URLs"
            )


def glob_directory(pattern: str) -> Generator[str, None, None]:
    """
    Similar to ``glob.glob()`` from the standard library, but works with remote directories as well.

    .. warning::
        Only a subset of glob patterns are supported. Specifically, ``*`` and ``**`` wildcards,
        which the follow the semantics defined here https://docs.python.org/3/library/pathlib.html#pattern-language.
    """
    # Pull out base directory from pattern by finding the first part before any wildcard.
    # Split by '/' and take path components until we hit one with a wildcard.
    parts = pattern.split("/")
    base_parts = []
    for part in parts:
        if "*" in part:
            break
        base_parts.append(part)
    dir = "/".join(base_parts) if base_parts else "."

    # Translate the glob pattern into a regex.
    # For example, "src/examples/**/*.py" --> "^src/examples/.*[^/]*\\.py$".
    pattern_regex = re.compile(
        "^"
        + re.escape(pattern).replace(r"\*\*/", ".*").replace(r"\*\*", ".*").replace(r"\*", "[^/]*")
        + "$"
    )

    for path in list_directory(dir, recurse="**" in pattern):
        if pattern_regex.match(path):
            yield path


def init_client(remote_path: str):
    """
    Initialize the right client for the given remote resource. This is helpful to avoid threading issues
    with boto3.
    """
    if remote_path.startswith("s3://"):
        _get_s3_client("s3")
    elif remote_path.startswith("r2://"):
        _get_s3_client("r2")
    elif remote_path.startswith("weka://"):
        _get_s3_client("weka")


###################################
## Serialization/deserialization ##
###################################


def serialize_to_tensor(x: Any) -> torch.Tensor:
    """
    Serialize an object to a byte tensor using pickle.

    :param x: The pickeable object to serialize.
    """
    serialized_bytes = pickle.dumps(x)
    return torch.frombuffer(bytearray(serialized_bytes), dtype=torch.uint8)


def deserialize_from_tensor(data: torch.Tensor) -> Any:
    """
    Deserialize an object from a byte tensor using pickle.

    :param data: The byte tensor to deserialize.
    """
    assert data.dtype == torch.uint8
    return pickle.loads(bytearray([int(x.item()) for x in data.flatten()]))


######################
## Internal helpers ##
######################


def _wait_before_retry(attempt: int):
    time.sleep(min(0.5 * 2**attempt, 3.0))


def _format_bytes(num: Union[int, float], suffix="B") -> str:
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


def retriable(
    max_attempts: int = 3,
    retriable_errors: Tuple[Type[Exception], ...] = (
        requests.exceptions.ConnectionError,
        requests.exceptions.Timeout,
    ),
    retry_condition: Optional[Callable[[Exception], bool]] = None,
):
    def decorator(func):
        @wraps(func)
        def new_func(*args, **kwargs):
            for attempt in range(1, max_attempts + 1):
                try:
                    return func(*args, **kwargs)
                except Exception as exc:
                    if isinstance(exc, retriable_errors) or (
                        retry_condition is not None and retry_condition(exc)
                    ):
                        if attempt >= max_attempts:
                            # When torch's DataLoader intercepts exceptions, it may try to re-raise them
                            # by recalling their constructor with a single message arg. Torch has some
                            # logic to deal with the absence of a single-parameter constructor, but it
                            # doesn't gracefully handle other possible failures in calling such a constructor
                            # This can cause an irrelevant exception (e.g. KeyError: 'error'), resulting
                            # in us losing the true exception info. To avoid this, we change the exception
                            # to a type that has a single-parameter constructor.
                            raise OLMoNetworkError(
                                f"'{func.__name__}' failed {max_attempts} attempts with: {exc}"
                            ) from exc
                        else:
                            log.warning(
                                f"'{func.__name__}' failed attempt {attempt} with retriable error: {exc}"
                            )
                            _wait_before_retry(attempt)
                    else:
                        raise

        return new_func

    return decorator


######################
## HTTPS IO helpers ##
######################


@retriable()
def _http_file_size(url: str) -> int:
    response = requests.head(url, allow_redirects=True)
    content_length = response.headers.get("content-length")
    assert content_length
    return int(content_length)


@retriable()
def _http_get_bytes_range(url: str, bytes_start: int, num_bytes: int) -> bytes:
    response = requests.get(
        url, headers={"Range": f"bytes={bytes_start}-{bytes_start + num_bytes - 1}"}
    )
    if response.status_code == 404:
        raise FileNotFoundError(url)

    response.raise_for_status()

    result = response.content
    # Some web servers silently ignore range requests and send everything
    assert len(result) == num_bytes, f"expected {num_bytes} bytes, got {len(result)}"

    return result


@retriable()
def _http_file_exists(url: str) -> bool:
    response = requests.head(url)
    if response.status_code == 404:
        return False

    response.raise_for_status()
    return True


####################
## GCS IO helpers ##
####################


@cache
def _get_gcs_client():
    from google.cloud import storage as gcs

    return gcs.Client()


def _gcs_is_retriable(exc: Exception) -> bool:
    from google.api_core.exceptions import BadRequest, GatewayTimeout
    from google.api_core.retry import if_transient_error
    from google.auth.exceptions import RefreshError

    return if_transient_error(exc) or isinstance(
        exc,
        (
            requests.exceptions.Timeout,
            BadRequest,  # Weird choice, but Google throws this transiently
            GatewayTimeout,
            RefreshError,
        ),
    )


def _get_gcs_retry():
    from google.api_core.retry import Retry

    return Retry(
        predicate=_gcs_is_retriable,  # NOTE: it appears google might ignore this
        initial=1.0,
        maximum=10.0,
        multiplier=2.0,
        timeout=600.0,
    )


@retriable(retry_condition=_gcs_is_retriable)
def _gcs_file_size(bucket_name: str, key: str) -> int:
    from google.api_core.exceptions import NotFound

    storage_client = _get_gcs_client()
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(key)
    try:
        blob.reload(retry=_get_gcs_retry())
    except NotFound:
        raise FileNotFoundError(f"gs://{bucket_name}/{key}")
    assert blob.size is not None
    return blob.size


@retriable(retry_condition=_gcs_is_retriable)
def _gcs_remove_file(bucket_name: str, key: str):
    from google.api_core.exceptions import NotFound

    storage_client = _get_gcs_client()
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(key)
    try:
        blob.reload(retry=_get_gcs_retry())
        bucket.delete_blob(blob.name)
    except NotFound:
        raise FileNotFoundError(f"gs://{bucket_name}/{key}")


@retriable(retry_condition=_gcs_is_retriable)
def _gcs_get_bytes_range(bucket_name: str, key: str, bytes_start: int, num_bytes: int) -> bytes:
    from google.api_core.exceptions import NotFound

    storage_client = _get_gcs_client()
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(key)
    try:
        blob.reload(retry=_get_gcs_retry())
    except NotFound:
        raise FileNotFoundError(f"gs://{bucket_name}/{key}")
    return blob.download_as_bytes(
        start=bytes_start,
        end=bytes_start + num_bytes - 1,
        retry=_get_gcs_retry(),
        checksum=None,  # type: ignore
    )


@retriable(retry_condition=_gcs_is_retriable)
def _gcs_upload(source: Path, bucket_name: str, key: str, save_overwrite: bool = False):
    storage_client = _get_gcs_client()
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(key)

    if blob.exists(retry=_get_gcs_retry()) and not save_overwrite:
        raise FileExistsError(
            f"gs://{bucket_name}/{key} already exists. Use save_overwrite to overwrite it."
        )

    try:
        blob.upload_from_filename(
            source,
            # NOTE: mypy and language servers may complain about the type here, but it does
            # not in fact need to be a ConditionalRetry, a plain old Retry is fine.
            retry=_get_gcs_retry(),  # type: ignore
        )
    except Exception as e:
        raise OLMoUploadError(
            f"Failed to upload '{source}' to '{key}' in GCS bucket '{bucket_name}'"
        ) from e


@retriable(retry_condition=_gcs_is_retriable)
def _gcs_clear_directory(bucket_name: str, prefix: str):
    from google.api_core.exceptions import NotFound

    storage_client = _get_gcs_client()

    prefix = prefix.strip("/")
    if prefix:
        prefix += "/"

    try:
        bucket = storage_client.bucket(bucket_name)
        blobs = bucket.list_blobs(prefix=prefix, retry=_get_gcs_retry())
        for blob in blobs:
            bucket.delete_blob(blob.name)
    except NotFound:
        return


def _gcs_list_directory(
    bucket_name: str,
    prefix: str,
    recurse: bool = False,
    include_files: bool = True,
    include_dirs: bool = True,
) -> Generator[str, None, None]:
    from google.api_core.exceptions import NotFound

    storage_client = _get_gcs_client()

    prefix = prefix.strip("/")
    if prefix:
        prefix += "/"

    for match_glob in (
        prefix + "*",  # only immediate files
        prefix + "**/",  # only immediate sub-folders
    ):
        try:
            bucket = storage_client.bucket(bucket_name)
            blobs = bucket.list_blobs(
                prefix=prefix,
                delimiter="/",
                match_glob=match_glob,
                retry=_get_gcs_retry(),
            )
        except NotFound:
            raise FileNotFoundError(f"gs://{bucket_name}/{prefix}")

        if include_files:
            for blob in blobs:
                yield f"gs://{bucket_name}/{blob.name}"

        for folder in blobs.prefixes:
            if include_dirs:
                yield f"gs://{bucket_name}/{folder.strip('/')}"
            if recurse:
                yield from _gcs_list_directory(
                    bucket_name,
                    folder,
                    recurse=True,
                    include_files=include_files,
                    include_dirs=include_dirs,
                )


###################
## S3 IO helpers ##
###################


@cache
def _get_s3_client(scheme: str):
    import boto3
    from botocore import UNSIGNED
    from botocore.config import Config

    session = boto3.Session(profile_name=_get_s3_profile_name(scheme))

    credentials = session.get_credentials()
    config = Config(retries={"max_attempts": 10, "mode": "standard"})
    if credentials is None:
        config = config.merge(Config(signature_version=UNSIGNED))

    return session.client(
        "s3",
        endpoint_url=_get_s3_endpoint_url(scheme),
        config=config,
        use_ssl=not int(os.environ.get("OLMO_NO_SSL", "0")),
    )


def _get_s3_profile_name(scheme: str) -> Optional[str]:
    if scheme == "s3":
        # For backwards compatibility, we assume S3 uses the default profile if S3_PROFILE is not set.
        return os.environ.get("S3_PROFILE")
    if scheme == "r2":
        profile_name = os.environ.get("R2_PROFILE")
        if profile_name is None:
            raise OLMoEnvironmentError(
                "R2 profile name is not set. Did you forget to set the 'R2_PROFILE' env var?"
            )

        return profile_name
    if scheme == "weka":
        profile_name = os.environ.get("WEKA_PROFILE")
        if profile_name is None:
            raise OLMoEnvironmentError(
                "WEKA profile name is not set. Did you forget to set the 'WEKA_PROFILE' env var?"
            )

        return profile_name

    raise NotImplementedError(f"Cannot get profile name for scheme {scheme}")


def _get_s3_endpoint_url(scheme: str) -> Optional[str]:
    if scheme == "s3":
        return None
    if scheme == "r2":
        r2_endpoint_url = os.environ.get("R2_ENDPOINT_URL")
        if r2_endpoint_url is None:
            raise OLMoEnvironmentError(
                "R2 endpoint url is not set. Did you forget to set the 'R2_ENDPOINT_URL' env var?"
            )

        return r2_endpoint_url
    if scheme == "weka":
        weka_endpoint_url = os.environ.get("WEKA_ENDPOINT_URL")
        if weka_endpoint_url is None:
            raise OLMoEnvironmentError(
                "WEKA endpoint url is not set. Did you forget to set the 'WEKA_ENDPOINT_URL' env var?"
            )

        return weka_endpoint_url

    raise NotImplementedError(f"Cannot get endpoint url for scheme {scheme}")


def _s3_retry_condition(err: Exception) -> bool:
    import botocore.exceptions as boto_errors

    return isinstance(
        err, (boto_errors.ClientError, boto_errors.HTTPClientError, boto_errors.ConnectionError)
    )


@retriable(retry_condition=_s3_retry_condition)
def _s3_file_size(scheme: str, bucket_name: str, key: str) -> int:
    from botocore.exceptions import ClientError

    try:
        return _get_s3_client(scheme).head_object(Bucket=bucket_name, Key=key)["ContentLength"]
    except ClientError as e:
        if e.response["ResponseMetadata"]["HTTPStatusCode"] == 404:
            raise FileNotFoundError(f"{scheme}://{bucket_name}/{key}") from e
        else:
            raise


@retriable(retry_condition=_s3_retry_condition)
def _s3_remove_file(scheme: str, bucket_name: str, key: str):
    from botocore.exceptions import ClientError

    try:
        return _get_s3_client(scheme).delete_object(Bucket=bucket_name, Key=key)
    except ClientError as e:
        if e.response["ResponseMetadata"]["HTTPStatusCode"] == 404:
            raise FileNotFoundError(f"{scheme}://{bucket_name}/{key}") from e
        else:
            raise


@retriable(retry_condition=_s3_retry_condition)
def _s3_get_bytes_range(
    scheme: str, bucket_name: str, key: str, bytes_start: int, num_bytes: int
) -> bytes:
    from botocore.exceptions import ClientError

    try:
        return (
            _get_s3_client(scheme)
            .get_object(
                Bucket=bucket_name,
                Key=key,
                Range=f"bytes={bytes_start}-{bytes_start + num_bytes - 1}",
            )["Body"]
            .read()
        )
    except ClientError as e:
        if e.response["ResponseMetadata"]["HTTPStatusCode"] == 404:
            raise FileNotFoundError(f"{scheme}://{bucket_name}/{key}") from e
        elif e.response["Error"]["Code"] == "InvalidRange":
            raise OLMoInvalidRangeRequestError(
                f"Invalid range request to '{scheme}://{bucket_name}/{key}' ({bytes_start=}, {num_bytes=})"
            )
        else:
            raise


@retriable(retry_condition=_s3_retry_condition)
def _s3_upload(
    source: Path,
    scheme: str,
    bucket_name: str,
    key: str,
    save_overwrite: bool = False,
):
    from botocore.exceptions import ClientError

    if not save_overwrite:
        try:
            _get_s3_client(scheme).head_object(Bucket=bucket_name, Key=key)
            raise FileExistsError(
                f"s3://{bucket_name}/{key} already exists. Use save_overwrite to overwrite it."
            )
        except ClientError as e:
            if e.response["ResponseMetadata"]["HTTPStatusCode"] == 404:
                pass
            else:
                raise

    _get_s3_client(scheme).upload_file(source, bucket_name, key)


@retriable(retry_condition=_s3_retry_condition)
def _s3_clear_directory(scheme: str, bucket_name: str, prefix: str):
    from botocore.exceptions import ClientError

    if not prefix.endswith("/"):
        prefix = prefix + "/"

    try:
        for o in _get_s3_client(scheme).list_objects_v2(Bucket=bucket_name, Prefix=prefix)[
            "Contents"
        ]:
            _get_s3_client(scheme).delete_object(Bucket=bucket_name, Key=o["Key"])
        return
    except ClientError as e:
        if e.response["ResponseMetadata"]["HTTPStatusCode"] == 404:
            return
    except KeyError:
        return


def _s3_list_directory(
    scheme: str,
    bucket_name: str,
    prefix: str,
    recurse: bool = False,
    include_files: bool = True,
    include_dirs: bool = True,
) -> Generator[str, None, None]:
    client = _get_s3_client(scheme)
    paginator = client.get_paginator("list_objects_v2")
    if not prefix.endswith("/"):
        prefix = prefix + "/"
    for page in paginator.paginate(Bucket=bucket_name, Prefix=prefix, MaxKeys=50, Delimiter="/"):
        if include_files:
            for file_item in page.get("Contents", []):
                yield f"{scheme}://{bucket_name}/{file_item['Key']}"
        for dir_item in page.get("CommonPrefixes", []):
            if include_dirs:
                yield f"{scheme}://{bucket_name}/{dir_item['Prefix'].strip('/')}"
            if recurse:
                yield from _s3_list_directory(
                    scheme,
                    bucket_name,
                    dir_item["Prefix"],
                    recurse=True,
                    include_files=include_files,
                    include_dirs=include_dirs,
                )


#############################################
## Custom cached path client for 'weka://' ##
#############################################


def add_cached_path_clients():
    """
    Add additional cached-path clients.
    """
    add_scheme_client(_WekaClient)


class _WekaClient(SchemeClient):
    recoverable_errors = S3Client.recoverable_errors

    scheme = "weka"

    def __init__(self, resource: str) -> None:
        super().__init__(resource)
        self.bucket_name, self.path = _WekaClient._split_cloud_path(resource, "weka")
        self.s3 = _get_s3_client("weka")
        self.object_info = None

    @staticmethod
    def _split_cloud_path(url: str, provider: str) -> Tuple[str, str]:
        """Split a full s3 path into the bucket name and path."""
        from urllib.parse import urlparse

        parsed = urlparse(url)
        if not parsed.netloc or not parsed.path:
            raise ValueError("bad {} path {}".format(provider, url))
        bucket_name = parsed.netloc
        provider_path = parsed.path
        # Remove '/' at beginning of path.
        if provider_path.startswith("/"):
            provider_path = provider_path[1:]
        return bucket_name, provider_path

    def _ensure_object_info(self):
        import botocore.exceptions as boto_exceptions

        if self.object_info is None:
            try:
                self.object_info = self.s3.head_object(Bucket=self.bucket_name, Key=self.path)
            except boto_exceptions.ClientError as e:
                if e.response["ResponseMetadata"]["HTTPStatusCode"] == 404:
                    raise FileNotFoundError(f"weka://{self.bucket_name}/{self.path}") from e
                raise e

    def get_etag(self) -> Optional[str]:
        self._ensure_object_info()
        assert self.object_info is not None
        return self.object_info.get("ETag")

    def get_size(self) -> Optional[int]:
        self._ensure_object_info()
        assert self.object_info is not None
        return self.object_info.get("ContentLength")

    def get_resource(self, temp_file: io.BufferedWriter) -> None:
        self.s3.download_fileobj(Fileobj=temp_file, Bucket=self.bucket_name, Key=self.path)

    def get_bytes_range(self, index: int, length: int) -> bytes:
        response = self.s3.get_object(
            Bucket=self.bucket_name, Key=self.path, Range=f"bytes={index}-{index + length - 1}"
        )
        return response["Body"].read()



"""
Common ``nn`` function implementations.
"""

import torch

from .cross_entropy_loss import *

__all__ = [
    "cross_entropy_loss",
    "fused_linear_cross_entropy_loss",
    "l2_normalize",
]


def l2_normalize(x: torch.Tensor, dim: int = -1) -> torch.Tensor:
    # NOTE: could also use F.normalize(), but that doesn't work with DTensor at the moment.
    return x / torch.linalg.vector_norm(x, dim=dim, keepdim=True, dtype=torch.float32).type_as(x)



import logging
import math
from dataclasses import dataclass
from typing import Any, Dict, Optional

from olmo_core.data import NumpyFSLDataLoader, NumpyFSLDataset
from olmo_core.data.utils import melt_batch, truncate_batch
from olmo_core.exceptions import OLMoConfigurationError
from olmo_core.utils import gc_cuda

from ..train_module import TransformerTrainModule
from .callback import Callback

log = logging.getLogger(__name__)


@dataclass
class SequenceLengthSchedulerCallback(Callback):
    """
    A :class:`Callback` for introducing a linear sequence-length warm-up schedule
    over the course of :data:`warmup_steps` starting from :data:`min_sequence_length`
    and ending at the configured training sequence length
    (:data:`NumpyFSLDataset.sequence_length <olmo_core.data.NumpyFSLDataset.sequence_length`).

    When :data:`truncate` is ``False`` the scheduler works by splitting each instance in a batch
    into more shorter instances while maintaining the same number of tokens in each batch and micro-batch.
    In this case the sequence length set during the warm-up will always be a multiple of
    :data:`min_sequence_length` by a power of 2, and therefore
    the train sequence length must be a multiple of :data:`min_sequence_length` by a power of 2.

    Otherwise the scheduler simply truncates the instances in the batch to the desired sequence
    length, throwing out the extra tokens. The scheduler will ensure the sequence length
    during the warm-up is always a multiple of :data:`keep_multiple_of`.

    .. important::
        This callback is only compatible with a :class:`~olmo_core.data.data_loader.NumpyFSLDataLoader`
        training :data:`~olmo_core.train.Trainer.data_loader`.

    .. note::
        The "total tokens" recorded by the trainer and :class:`SpeedMonitorCallback` will
        still include tokens truncated by this callback for bookkeeping purposes.
    """

    min_sequence_length: int = 128
    warmup_steps: int = 2000
    truncate: bool = False
    keep_multiple_of: int = 128
    enabled: bool = True

    _og_rank_microbatch_size: Optional[int] = None
    _last_seq_len: Optional[int] = None

    def pre_train(self):
        if not self.enabled:
            return

        if not isinstance(self.trainer.data_loader, NumpyFSLDataLoader):
            raise OLMoConfigurationError(
                "The sequence length scheduler callback requires a 'NumpyFSLDataLoader', "
                f"got '{type(self.trainer.data_loader)}' instead"
            )
        if not isinstance(self.trainer.train_module, TransformerTrainModule):
            raise OLMoConfigurationError(
                "The sequence length scheduler callback requires a 'TransformerTrainModule', "
                f"got '{type(self.trainer.train_module)}' instead"
            )

        dataset = self.trainer.data_loader.dataset
        assert isinstance(dataset, NumpyFSLDataset)

        if self.truncate and (
            dataset.sequence_length % self.min_sequence_length != 0
            or (math.log(dataset.sequence_length // self.min_sequence_length, 2) % 1 != 0)
        ):
            raise OLMoConfigurationError(
                "train sequence length must be a multiple of 'min_sequence_length' by a power of 2 "
                "when 'truncate=False'."
            )
        elif dataset.sequence_length <= self.min_sequence_length:
            raise OLMoConfigurationError(
                "train sequence length must be greater than 'min_sequence_length'"
            )

        self._og_rank_microbatch_size = self.trainer.train_module.rank_microbatch_size

    def pre_step(self, batch: Dict[str, Any]):
        if not self.enabled:
            return

        if self.step > self.warmup_steps:
            return

        assert isinstance(self.trainer.data_loader, NumpyFSLDataLoader)
        dataset = self.trainer.data_loader.dataset
        assert isinstance(dataset, NumpyFSLDataset)
        assert isinstance(self.trainer.train_module, TransformerTrainModule)

        new_seq_len: int
        if self.truncate:
            new_seq_len = _get_truncated_sequence_length(
                self.min_sequence_length,
                dataset.sequence_length,
                self.step,
                self.warmup_steps,
                self.keep_multiple_of,
            )

            for key, value in truncate_batch(
                batch,
                new_seq_len,
            ).items():
                batch[key] = value
        else:
            new_seq_len = _get_split_sequence_length(
                self.min_sequence_length,
                dataset.sequence_length,
                self.step,
                self.warmup_steps,
            )

            for key, value in melt_batch(
                batch,
                new_seq_len,
            ).items():
                batch[key] = value

            # Increase micro-batch size proportionally to maintain the same number of tokens
            # in each micro-batch.
            assert self._og_rank_microbatch_size is not None
            new_rank_microbatch_size = self._og_rank_microbatch_size * (
                dataset.sequence_length // new_seq_len
            )
            self.trainer.train_module.rank_microbatch_size = new_rank_microbatch_size

        if new_seq_len != self._last_seq_len:
            log.info(f"Changing sequence length to {new_seq_len} per warm-up schedule")
            self._last_seq_len = new_seq_len
            # Empty CUDA cache since shapes have now changed.
            gc_cuda()

    def post_train_batch(self):
        if not self.enabled or self.step > self.warmup_steps + 1:
            return

        assert isinstance(self.trainer.train_module, TransformerTrainModule)
        assert self._og_rank_microbatch_size is not None
        self.trainer.train_module.rank_microbatch_size = self._og_rank_microbatch_size


def _get_split_sequence_length(
    min_sequence_length: int, max_sequence_length: int, step: int, warmup_steps: int
) -> int:
    seq_len = (
        min_sequence_length
        + (max_sequence_length - min_sequence_length) * min(step, warmup_steps) / warmup_steps
    )

    n = math.floor(math.log(seq_len // min_sequence_length, 2))
    return min_sequence_length * 2**n


def _get_truncated_sequence_length(
    min_sequence_length: int,
    max_sequence_length: int,
    step: int,
    warmup_steps: int,
    keep_multiple_of: int,
) -> int:
    seq_len = (
        min_sequence_length
        + (max_sequence_length - min_sequence_length) * min(step, warmup_steps) / warmup_steps
    )

    seq_len = keep_multiple_of * (seq_len // keep_multiple_of)
    return int(seq_len)



_MAJOR = "2"
_MINOR = "3"
_PATCH = "0"
_SUFFIX = ""

VERSION_SHORT = "{0}.{1}".format(_MAJOR, _MINOR)
VERSION = "{0}.{1}.{2}{3}".format(_MAJOR, _MINOR, _PATCH, _SUFFIX)


if __name__ == "__main__":
    import sys

    if sys.argv[-1] == "short":
        print(VERSION_SHORT)
    else:
        print(VERSION)



"""
Dataset, data loaders, and config builders for use with the :class:`~olmo_core.train.Trainer`.

Overview
--------

For text-based data you should prepare your data by writing token IDs to numpy arrays on disk, using the
`Dolma toolkit <https://allenai.github.io/dolma/>`_ for example.
Then configure and build your dataset using one of the
:class:`~olmo_core.data.numpy_dataset.NumpyDatasetConfigBase` builders (for example
``NumpyFSLDatasetConfig``), build your data loader with the
:class:`~olmo_core.data.data_loader.NumpyDataLoaderConfig`
builder, and pass it to :meth:`TrainerConfig.build() <olmo_core.train.TrainerConfig.build>`.
"""

from .collator import DataCollator, PaddingDirection
from .data_loader import (
    DataLoaderBase,
    NumpyDataLoaderBase,
    NumpyDataLoaderConfig,
    NumpyFSLDataLoader,
    NumpyVSLDataLoader,
    TextDataLoaderBase,
)
from .mixes import DataMix, DataMixBase
from .numpy_dataset import (
    InstanceFilterConfig,
    NumpyDatasetBase,
    NumpyDatasetConfig,
    NumpyFSLDataset,
    NumpyFSLDatasetBase,
    NumpyFSLDatasetConfig,
    NumpyInterleavedFSLDatasetConfig,
    NumpyPackedFSLDataset,
    NumpyPackedFSLDatasetConfig,
    NumpyPaddedFSLDataset,
    NumpyPaddedFSLDatasetConfig,
    NumpyVSLDataset,
    NumpyVSLDatasetConfig,
    VSLCurriculum,
    VSLCurriculumConfig,
    VSLCurriculumType,
    VSLGrowLinearCurriculum,
    VSLGrowP2Curriculum,
    VSLGrowthCurriculum,
    VSLNaturalCurriculum,
)
from .tokenizer import TokenizerConfig, TokenizerName
from .types import LongDocStrategy, NumpyDatasetDType

__all__ = [
    "NumpyDatasetBase",
    "NumpyFSLDatasetBase",
    "NumpyFSLDataset",
    "NumpyPaddedFSLDataset",
    "NumpyPackedFSLDataset",
    "NumpyVSLDataset",
    "VSLCurriculum",
    "VSLNaturalCurriculum",
    "VSLGrowthCurriculum",
    "VSLGrowP2Curriculum",
    "VSLGrowLinearCurriculum",
    "NumpyDatasetConfig",
    "NumpyFSLDatasetConfig",
    "NumpyPaddedFSLDatasetConfig",
    "NumpyPackedFSLDatasetConfig",
    "NumpyInterleavedFSLDatasetConfig",
    "NumpyVSLDatasetConfig",
    "InstanceFilterConfig",
    "VSLCurriculumType",
    "VSLCurriculumConfig",
    "NumpyDatasetDType",
    "TokenizerConfig",
    "TokenizerName",
    "DataMixBase",
    "DataMix",
    "DataCollator",
    "PaddingDirection",
    "DataLoaderBase",
    "TextDataLoaderBase",
    "NumpyDataLoaderBase",
    "NumpyFSLDataLoader",
    "NumpyVSLDataLoader",
    "NumpyDataLoaderConfig",
    "LongDocStrategy",
]



import os
from dataclasses import dataclass
from typing import Optional, Tuple

import requests

from olmo_core.config import Config

GIT_REPO_URL_ENV_VAR = "REPO_URL"
GIT_REF_ENV_VAR = "GIT_REF"
GIT_BRANCH_ENV_VAR = "GIT_BRANCH"


def parse_git_remote_url(url: str) -> Tuple[str, str]:
    """
    Parse a git remote URL into a GitHub (account, repo) pair.

    :raises InvalidRemoteError: If the URL can't be parsed correctly.
    """
    if "github.com" not in url:
        raise ValueError(f"Remote ('{url}') must point to a GitHub repo")
    try:
        account, repo = url.split("github.com", 1)[-1].strip("/:").split(".git")[0].split("/")
    except ValueError:
        raise ValueError(f"Failed to parse GitHub repo path from remote '{url}'")
    return account, repo


@dataclass
class GitConfig(Config):
    repo_url: str
    ref: str
    branch: Optional[str] = None

    @property
    def is_dirty(self) -> bool:
        from git.exc import InvalidGitRepositoryError
        from git.repo import Repo

        try:
            repo = Repo(".")
            return repo.is_dirty()
        except InvalidGitRepositoryError:
            return False

    @property
    def is_public(self) -> bool:
        response = requests.get(self.repo_url)
        if response.status_code not in {200, 404}:
            response.raise_for_status()
        return response.status_code == 200

    @classmethod
    def from_env(cls) -> Optional["GitConfig"]:
        from git.exc import InvalidGitRepositoryError
        from git.repo import Repo

        try:
            repo = Repo(".")
        except InvalidGitRepositoryError:
            return None

        git_ref = os.environ.get(GIT_REF_ENV_VAR, str(repo.commit()))
        remote = repo.remote()

        # Try to find a remote based on the current tracking branch.
        try:
            branch = repo.active_branch
        except TypeError:
            branch = None

        if branch is not None:
            branch = branch.tracking_branch()

        branch_name = os.environ.get(GIT_BRANCH_ENV_VAR)
        if branch is not None:
            remote = repo.remote(branch.remote_name)
            if branch_name is None:
                assert branch.name.startswith(branch.remote_name + "/")
                branch_name = branch.name.replace(branch.remote_name + "/", "", 1)

        if (repo_url := os.environ.get(GIT_REPO_URL_ENV_VAR)) is None:
            account, repo_name = parse_git_remote_url(remote.url)
            repo_url = f"https://github.com/{account}/{repo_name}"

        return cls(
            repo_url=repo_url,
            ref=git_ref,
            branch=branch_name,
        )



"""
Distributed helpers, most of which work in a non-distributed context as well for API unity.
"""

import logging
import math
import os
from datetime import timedelta
from typing import Callable, List, Optional, TypeVar, Union, cast

import torch
import torch.distributed as dist
from torch.distributed.device_mesh import DeviceMesh
from torch.distributed.tensor import DTensor, distribute_tensor

from ..exceptions import OLMoEnvironmentError
from ..utils import log_or_print, move_to_device, set_env_var

OLMO_SHARED_FS_ENV_VAR = "OLMO_SHARED_FS"
OLMO_FS_LOCAL_RANK_ENV_VAR = "FS_LOCAL_RANK"
OLMO_LOCAL_RANK_ENV_VAR = "LOCAL_RANK"
OLMO_NUM_NODES_ENV_VAR = "NUM_NODES"
OLMO_LOCAL_WORLD_SIZE_ENV_VAR = "LOCAL_WORLD_SIZE"
BEAKER_HOSTNAME_ENV_VAR = "BEAKER_NODE_HOSTNAME"


log = logging.getLogger(__name__)


def init_distributed(
    backend: str = "nccl",
    timeout: timedelta = timedelta(minutes=30),
    shared_filesytem: Optional[bool] = True,
    **kwargs,
):
    """
    Initialize the distributed process group with the given backend(s) and check/set the
    relevant environment variables.
    This also calls :func:`torch.cuda.set_device()` for backends that support CUDA.
    """
    # To mitigate the memory issue that collectives using async_op=True hold memory longer
    # than they should such as those in tensor parallelism.
    set_env_var("TORCH_NCCL_AVOID_RECORD_STREAMS", "1")

    # Force processes to synchronize at init process group.
    set_env_var("TORCH_DIST_INIT_BARRIER", "1")

    if shared_filesytem:
        set_env_var(OLMO_SHARED_FS_ENV_VAR, "1")

    # Set host-specific env var defaults.
    if _running_in_beaker():
        node_name = get_node_hostname()
        log_or_print(log, f"Running in beaker on node '{node_name}'")
        multi_node = int(os.environ.get(OLMO_NUM_NODES_ENV_VAR, "1")) > 1

        # See https://beaker-docs.apps.allenai.org/experiments/distributed-training.html
        if "jupiter" in node_name:
            set_env_var("NCCL_IB_HCA", "^=mlx5_bond_0")
            if multi_node:
                # Only for multi-node
                set_env_var("NCCL_SOCKET_IFNAME", "ib")
        elif "pluto" in node_name:
            set_env_var("NCCL_IB_HCA", "^=mlx5_1,mlx5_2")
        elif "augusta" in node_name:
            # NOTE: For single-node training we still need all of these settings and we also
            # need host networking enabled so that the ethernet interface names don't change.
            set_env_var("NCCL_CROSS_NIC", "0")
            set_env_var("NCCL_ALGO", "Ring,Tree")
            set_env_var("NCCL_PROTO", "Simple,LL128")
            set_env_var("NCCL_MIN_NCHANNELS", "4")
            set_env_var("NCCL_P2P_NET_CHUNKSIZE", "524288")
            set_env_var("NCCL_P2P_PCI_CHUNKSIZE", "524288")
            set_env_var("NCCL_P2P_NVL_CHUNKSIZE", "1048576")
            set_env_var("NCCL_FASTRAK_NUM_FLOWS", "2")
            set_env_var("NCCL_FASTRAK_ENABLE_CONTROL_CHANNEL", "0")
            set_env_var("NCCL_BUFFSIZE", "8388608")
            set_env_var("NCCL_FASTRAK_USE_SNAP", "1")
            set_env_var("CUDA_VISIBLE_DEVICES", "0,1,2,3,4,5,6,7")
            set_env_var("NCCL_NET_GDR_LEVEL", "PIX")
            set_env_var("NCCL_FASTRAK_ENABLE_HOTPATH_LOGGING", "0")
            set_env_var(
                "NCCL_FASTRAK_PLUGIN_ACCEPT_TIMEOUT_MS", str(int(timeout.total_seconds() * 1000))
            )
            set_env_var("NCCL_NVLS_ENABLE", "0")
            set_env_var("NCCL_USE_SNAP", "1")
            set_env_var("NCCL_FASTRAK_USE_LLCM", "1")
            set_env_var("NCCL_FASTRAK_LLCM_DEVICE_DIRECTORY", "/dev/aperture_devices")
            # NOTE: This path var must be set prior to launching Python
            #  set_env_var(
            #      "LD_LIBRARY_PATH",
            #      "/var/lib/tcpxo/lib64:" + os.environ.get("LD_LIBRARY_PATH", ""),
            #      override=True,
            #  )
            set_env_var("NCCL_TUNER_PLUGIN", "libnccl-tuner.so")
            set_env_var(
                "NCCL_TUNER_CONFIG_PATH", "/var/lib/tcpxo/lib64/a3plus_tuner_config_ll128.textproto"
            )
            set_env_var(
                "NCCL_SHIMNET_GUEST_CONFIG_CHECKER_CONFIG_FILE",
                "/var/lib/tcpxo/lib64/a3plus_guest_config_ll128.textproto",
            )
            set_env_var("NCCL_FASTRAK_CTRL_DEV", "enp0s12")
            set_env_var(
                "NCCL_FASTRAK_IFNAME",
                "enp6s0,enp7s0,enp13s0,enp14s0,enp134s0,enp135s0,enp141s0,enp142s0",
            )
            set_env_var("NCCL_SOCKET_IFNAME", "enp0s12")
            set_env_var("NCCL_DEBUG_SUBSYS", "INIT,NET")

    if backend_supports_cuda(backend):
        # Set CUDA device.
        # NOTE: important to do this *before* initializing the process group to avoid
        # other ranks initializing CUDA on GPU 0.
        device = torch.device(f"cuda:{int(os.environ[OLMO_LOCAL_RANK_ENV_VAR])}")
        torch.cuda.set_device(device)

    dist.init_process_group(backend, timeout=timeout, **kwargs)

    validate_env_vars()

    log_or_print(
        log,
        f"Global rank {get_rank()} "
        f"= local rank {get_local_rank()} "
        f"= file system local rank {get_fs_local_rank()}",
    )


def validate_env_vars():
    """
    Validate distributed environment variables. This is called internally by :func:`init_distributed()`.
    """
    if not is_distributed():
        return

    if OLMO_LOCAL_RANK_ENV_VAR not in os.environ:
        raise OLMoEnvironmentError(f"Missing env var '{OLMO_LOCAL_RANK_ENV_VAR}'")

    if os.environ.get(OLMO_SHARED_FS_ENV_VAR) != "1" and (
        os.environ.get(OLMO_FS_LOCAL_RANK_ENV_VAR) is None
        and os.environ.get(OLMO_LOCAL_RANK_ENV_VAR) is None
    ):
        raise OLMoEnvironmentError(
            f"Missing env var '{OLMO_FS_LOCAL_RANK_ENV_VAR}'/'{OLMO_LOCAL_RANK_ENV_VAR}' for non-shared filesystem. "
            f"If this is a shared filesystem you can set '{OLMO_SHARED_FS_ENV_VAR}=1' instead."
        )

    if OLMO_NUM_NODES_ENV_VAR not in os.environ and OLMO_LOCAL_WORLD_SIZE_ENV_VAR not in os.environ:
        raise OLMoEnvironmentError(
            f"Missing either '{OLMO_NUM_NODES_ENV_VAR}' or '{OLMO_LOCAL_WORLD_SIZE_ENV_VAR}' env vars"
        )


def _running_in_beaker() -> bool:
    return BEAKER_HOSTNAME_ENV_VAR in os.environ


def get_node_hostname() -> str:
    """
    Get the hostname of the node.
    """
    if BEAKER_HOSTNAME_ENV_VAR in os.environ:
        return os.environ[BEAKER_HOSTNAME_ENV_VAR]
    else:
        import socket

        return socket.gethostname()


def is_distributed() -> bool:
    """
    Check if in a distributed context.
    """
    return dist.is_available() and dist.is_initialized()


def barrier(group: Optional[dist.ProcessGroup] = None) -> None:
    """
    Wait for all ranks in the group.
    """
    if is_distributed():
        dist.barrier(group)


def get_rank(group: Optional[dist.ProcessGroup] = None) -> int:
    """
    Get the rank within the process group.
    """
    if is_distributed():
        return dist.get_rank(group)
    else:
        return 0


def get_global_rank(group_rank: int, group: Optional[dist.ProcessGroup] = None) -> int:
    """
    Translate a rank within a group into it's global rank.
    """
    if group is None or not is_distributed():
        return group_rank
    else:
        return dist.get_global_rank(group, group_rank)


def get_local_rank() -> int:
    """
    Get the local rank within the current node.

    .. warning::
        This relies on the environment variable ``LOCAL_RANK`` being set correctly, but
        will always return 0 if a distributed process group has not been initialized.

    :returns: The rank.
    """
    if is_distributed():
        return int(os.environ.get(OLMO_LOCAL_RANK_ENV_VAR) or 0)
    else:
        return 0


def get_fs_local_rank(group: Optional[dist.ProcessGroup] = None) -> int:
    """
    Get the local rank per filesystem, meaning that, regardless of the number of nodes,
    if all ranks share the same filesystem then :func:`get_fs_local_rank()` will be equivalent
    to :func:`get_rank()`, but if nodes do not share the same filesystem then
    :func:`get_fs_local_rank()` will be equivalent to :func:`get_local_rank()`.

    .. warning::
        This relies on some environment variables to determine the correct rank.
        If you are using a shared filesystem across nodes, you can simply set the environment
        variable ``OLMO_SHARED_FS=1``. Otherwise you can set ``FS_LOCAL_RANK`` for each process.

        This will always return 0 if a distributed process group has not been initialized.

    :returns: The rank.
    """
    if not is_distributed():
        return 0
    elif os.environ.get(OLMO_SHARED_FS_ENV_VAR) == "1":
        return int(os.environ.get(OLMO_FS_LOCAL_RANK_ENV_VAR) or get_rank(group))
    else:
        return int(os.environ.get(OLMO_FS_LOCAL_RANK_ENV_VAR) or get_local_rank())


def get_world_size(group: Optional[dist.ProcessGroup] = None) -> int:
    """
    Get the world size of the default distributed process group.

    .. warning::
        This will always return 1 if a distributed group has not been initialized.
    """
    if is_distributed():
        return dist.get_world_size(group)
    else:
        return 1


def get_local_world_size() -> int:
    """
    Get the local world size within the default distributed process group.

    .. warning::
        This relies on the 'LOCAL_WORLD_SIZE' env var but will always return 1 if a distributed
        process group has not been initialized.

    :returns: The local world size.
    """
    if not is_distributed():
        return 1
    else:
        return int(os.environ[OLMO_LOCAL_WORLD_SIZE_ENV_VAR])


def get_num_nodes() -> int:
    """
    Get the number of nodes in the default distributed process group.

    .. warning::
        This relies on either the 'NUM_NODES' or 'LOCAL_WORLD_SIZE' env var, but will always
        return 1 if a distributed process group has not been initialized.

    :returns: The number of nodes.
    """
    if not is_distributed():
        return 1
    elif OLMO_NUM_NODES_ENV_VAR in os.environ:
        return int(os.environ[OLMO_NUM_NODES_ENV_VAR])
    else:
        return get_world_size() // get_local_world_size()


V = TypeVar("V", bool, int, float, torch.Tensor)


def synchronize_value(
    value: V, device: torch.device, src: int = 0, group: Optional[dist.ProcessGroup] = None
) -> V:
    """
    Synchronize a value across the distributed process group.
    """
    if dist.is_available() and dist.is_initialized():
        is_tensor = isinstance(value, torch.Tensor)
        value_tensor = (
            move_to_device(value, device)
            if is_tensor
            else move_to_device(torch.tensor(value), device)
        )  # type: ignore
        dist.broadcast(value_tensor, src, group=group)
        return value_tensor if is_tensor else value_tensor.item()  # type: ignore
    else:
        return value


def synchronize_flag(
    flag: bool, device: torch.device, group: Optional[dist.ProcessGroup] = None
) -> bool:
    """
    Synchronize a boolean across the distributed process group.
    """
    return synchronize_value(flag, device, group=group)


def all_reduce_value(
    value: V, device: torch.device, op=dist.ReduceOp.SUM, group: Optional[dist.ProcessGroup] = None
) -> V:
    """
    All reduce a value across the distributed process group.
    """
    if dist.is_available() and dist.is_initialized():
        is_tensor = isinstance(value, torch.Tensor)
        value_tensor = (
            move_to_device(value, device)
            if is_tensor
            else move_to_device(torch.tensor(value), device)
        )  # type: ignore
        dist.all_reduce(value_tensor, op=op, group=group)
        return value_tensor if is_tensor else value_tensor.item()  # type: ignore
    else:
        return value


T = TypeVar("T")


def scatter_object(obj: T, src: int = 0, group: Optional[dist.ProcessGroup] = None) -> T:
    """
    Scatter an object using pickle to all ranks in the process group.
    """
    if not is_distributed():
        return obj

    output_list: List[T] = [obj]
    input_list = [obj] * get_world_size(group) if get_rank(group) == src else None
    dist.scatter_object_list(output_list, input_list, src=src, group=group)
    return output_list[0]


def all_gather(
    tensor: torch.Tensor, group: Optional[dist.ProcessGroup] = None
) -> List[torch.Tensor]:
    """
    All-gather tensors from the whole group into a list.
    """
    if not is_distributed():
        return [tensor]

    shapes = all_gather_object(tensor.shape, group=group)
    output_list = [
        move_to_device(torch.zeros(shape, dtype=tensor.dtype), tensor.device) for shape in shapes
    ]
    dist.all_gather(output_list, tensor, group=group)
    return output_list


def all_gather_object(obj: T, group: Optional[dist.ProcessGroup] = None) -> List[T]:
    """
    All-gather an object using pickle to all ranks in a process group.
    """
    if not is_distributed():
        return [obj]

    output_list = [obj] * get_world_size(group)
    dist.all_gather_object(output_list, obj, group=group)
    return output_list


def get_mesh_coordinates(mesh: "DeviceMesh", rank: Optional[int] = None) -> Optional[List[int]]:
    """
    Calculate the coordinates of a global rank on a device mesh.

    :param mesh: The device mesh.
    :param rank: The global rank. If ``None``, the current global rank is used.

    :return: The coordinates or ``None`` if the rank is not part of the mesh.
    """
    rank = rank if rank is not None else get_rank()
    rank_coords = (mesh.mesh == rank).nonzero()
    assert rank_coords.size(0) in (0, 1)
    return rank_coords[0].tolist() if rank_coords.size(0) > 0 else None


def backend_supports_cuda(backend: Optional[str] = None) -> bool:
    """
    Check if a distributed backend supports CUDA tensors.
    """
    if backend is None and not is_distributed():
        return torch.cuda.is_available()

    backend = backend or dist.get_backend()
    if "nccl" in backend:
        return True
    else:
        return False


def backend_supports_cpu(backend: Optional[str] = None) -> bool:
    """
    Check if a distributed backend supports CPU tensors.
    """
    if backend is None and not is_distributed():
        return True

    backend = backend or dist.get_backend()
    if "gloo" in backend or "mpi" in backend:
        return True
    else:
        return False


def get_reduce_divide_factor(world_size: int) -> float:
    factor: int = 1
    while world_size % factor == 0 and world_size / factor > factor:
        factor *= 2
    return float(factor)


def get_local_tensor(x: torch.Tensor) -> torch.Tensor:
    if isinstance(x, DTensor):
        x = x.to_local()
        # An `AsyncCollectiveTensor` might be returned, which means the local tensor is not ready
        # yet (i.e. communication is not finished). In this case we need to call `.wait()`
        # to wait the local tensor to be ready.
        if hasattr(x, "wait"):
            return x.wait()  # type: ignore
        else:
            return x
    else:
        return x


def get_full_tensor(x: torch.Tensor) -> torch.Tensor:
    if isinstance(x, DTensor):
        return x.full_tensor()
    else:
        return x


def distribute_like(source: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    if not isinstance(source, DTensor):
        return get_full_tensor(target)

    if isinstance(target, DTensor):
        if target.device_mesh == source.device_mesh and target.placements == source.placements:
            return target
        else:
            return target.redistribute(device_mesh=source.device_mesh, placements=source.placements)

    return distribute_tensor(target, device_mesh=source.device_mesh, placements=source.placements)


def do_n_at_a_time(
    f: Callable[[], T],
    *,
    n: Optional[int] = None,
    process_group: Optional[dist.ProcessGroup] = None,
    world_size: Optional[int] = None,
    local_rank: Optional[int] = None,
) -> T:
    """
    Call a function ``f`` in a distributed context from at most ``n`` ranks at a time.

    All ranks will eventually call the given function exactly once, at which point this function
    will return.

    :param f: The function to call from each rank.
    :param n: The level of concurrency, i.e. how many ranks are allowed to call ``f`` at once.
        This defaults to the number of nodes, in which case one rank from each node will
        call ``f`` at a time.
    :param process_group: The process group to use.
    """
    world_size = world_size if world_size is not None else get_world_size(process_group)
    local_rank = local_rank if local_rank is not None else get_rank(process_group)
    n = n if n is not None else get_num_nodes()
    group_count = math.ceil(world_size / n)
    group_rank = local_rank % group_count
    result: Optional[T] = None
    for active_group in range(group_count):
        if group_rank == active_group:
            result = f()
        barrier(process_group)
    return cast(T, result)


class _HiddenTensor:
    def __init__(self, x: torch.Tensor):
        self.x = x

    @property
    def device(self) -> torch.device:
        return self.x.device

    def to(self, *args, **kwargs) -> "_HiddenTensor":
        return _HiddenTensor(self.x.to(*args, **kwargs))


def hide_from_torch(x: torch.Tensor) -> _HiddenTensor:
    return _HiddenTensor(x)


def unhide_from_torch(x: Union[torch.Tensor, _HiddenTensor]) -> torch.Tensor:
    if isinstance(x, _HiddenTensor):
        return x.x
    else:
        return x



import argparse
import logging
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import google.auth
from beaker import Job, Priority
from google.cloud.compute_v1.services.instances.client import InstancesClient

from olmo_core.exceptions import BeakerInsufficientResourcesError

log = logging.getLogger(__name__)


@dataclass
class HostMetadata:
    block: str
    """
    The ID of the block in which the running instance is located. Instances within the same block
    experience low network latency.
    """

    subblock: str
    """
    The ID of the sub-block in which the running instance is located. Instances in the same sub-block
    experience lower network latency than instances in the same block.
    """

    machine: str
    """
    The ID of the machine ('host' according to Google) on which the running instance is located.
    Instances on the same machine experience the lowest possible network latency.
    """


def get_hosts_metadata_from_gcp(
    zone: str, *, credentials_path: Optional[Path] = None
) -> dict[str, HostMetadata]:
    if credentials_path:
        credentials, project_id = google.auth.load_credentials_from_file(str(credentials_path))
    else:
        credentials, project_id = google.auth.default()

    if project_id != "h100-cluster-owner":
        raise RuntimeError(
            f"Expected credentials for 'h100-cluster-owner', got credentials for {project_id}"
        )

    client = InstancesClient(credentials=credentials)
    instance_pages = client.list(project=project_id, zone=zone).pages
    return {
        f"{instance.name}.reviz.ai2.in": HostMetadata(
            block=instance.resource_status.physical_host_topology.block,
            subblock=instance.resource_status.physical_host_topology.subblock,
            machine=instance.resource_status.physical_host_topology.host,
        )
        for page in instance_pages
        for instance in page.items
    }


def get_occupied_beaker_hosts(
    hosts_metadata: dict[str, HostMetadata], beaker_cluster: str, beaker_priority: Priority
) -> set[str]:
    from olmo_core.internal.common import get_beaker_client

    beaker = get_beaker_client()
    assert beaker is not None

    occupied_hosts = set()

    cluster = beaker.cluster.get(beaker_cluster)
    nodes = beaker.cluster.nodes(cluster)
    for node in sorted(nodes, key=lambda node: node.hostname):
        host = node.hostname
        assert host not in occupied_hosts, f"Host {host} is somehow already in occupied hosts"
        if host not in hosts_metadata:
            log.warning(f"No metadata found for beaker host {host}")
            continue

        if node.cordoned is not None:
            # Treat cordoned node as occupied since it might be uncordoned later.
            occupied_hosts.add(host)
            continue

        jobs = beaker.job.list(node=node)
        for job in jobs:
            if (
                job.is_running
                and job.execution is not None
                and (resources := job.execution.spec.resources) is not None
                and resources.gpu_count is not None
                and resources.gpu_count > 0
                and not _is_job_preemptible(job, beaker_priority)
            ):
                occupied_hosts.add(host)
                break

    return occupied_hosts


def _is_job_preemptible(job: Job, desired_priority: Priority) -> bool:
    if not job.is_preemptible:
        return False

    assert job.priority is not None
    # Priorities are sorted highest to lowest by default
    sorted_priorities = list(Priority)
    return sorted_priorities.index(desired_priority) < sorted_priorities.index(job.priority)


def get_hostname_constraints(
    hosts_metadata: dict[str, HostMetadata],
    num_execution_units: int,
    num_hosts_per_task: int,
    num_tasks: int,
    occupied_hosts: Optional[set[str]] = None,
) -> list[list[str]]:
    if num_hosts_per_task % num_execution_units != 0:
        raise ValueError(
            "Number of execution units must be a divisor of number of hosts in a task "
            "(since a task runs one or more execution units)."
        )
    num_hosts_per_exec_unit = (num_tasks * num_hosts_per_task) // num_execution_units

    occupied_hosts = occupied_hosts or set()

    hosts_by_block: defaultdict[str, list[str]] = defaultdict(list)
    available_hosts_by_block: defaultdict[str, list[str]] = defaultdict(list)
    for host, metadata in hosts_metadata.items():
        hosts_by_block[metadata.block].append(host)
        if host not in occupied_hosts:
            available_hosts_by_block[metadata.block].append(host)

    # Sort blocks in descending order of number of available hosts
    sorted_blocks = sorted(
        available_hosts_by_block.keys(),
        key=lambda block: len(available_hosts_by_block[block]),
        reverse=True,
    )

    if len(available_hosts_by_block[sorted_blocks[0]]) > num_hosts_per_task * num_tasks:
        # If all the tasks can be fulfilled in a single block, let's do that!
        # Moreover, include occupied hosts since they might free up later.
        return [list(hosts_by_block[sorted_blocks[0]]) for _ in range(num_tasks)]

    hosts_per_task: list[list[str]] = []
    block_idx = 0
    for _ in range(num_tasks):
        task_hosts: list[str] = []

        while block_idx < len(sorted_blocks) and len(task_hosts) < num_hosts_per_task:
            block = sorted_blocks[block_idx]
            num_block_hosts = len(available_hosts_by_block[block])

            if num_block_hosts < num_hosts_per_exec_unit:
                # We have exhausted this block, move onto the next one
                block_idx += 1
                continue

            needed_hosts = num_hosts_per_task - len(task_hosts)
            assert needed_hosts % num_hosts_per_exec_unit == 0

            # Take the nearest multiple of num_hosts_per_exec_unit as the number of hosts to give to the current task
            host_count_from_block = min(
                needed_hosts,
                num_block_hosts // num_hosts_per_exec_unit * num_hosts_per_exec_unit,
            )
            assert host_count_from_block > 0

            block_hosts = available_hosts_by_block[block]
            task_hosts.extend(block_hosts[:host_count_from_block])

            available_hosts_by_block[block] = block_hosts[host_count_from_block:]

        hosts_per_task.append(task_hosts)

    if len(hosts_per_task) < num_tasks:
        raise BeakerInsufficientResourcesError(
            f"Could only satisfy {len(hosts_per_task)} out of {num_tasks} tasks"
        )
    if len(hosts_per_task[-1]) < num_hosts_per_task:
        raise BeakerInsufficientResourcesError(
            f"Could not satisfy task number {len(hosts_per_task) - 1}, only got {len(hosts_per_task[-1])} hosts"
        )

    return hosts_per_task


def get_beaker_hostname_constraints(
    num_nodes: int,
    num_execution_units: int,
    beaker_task_count: int,
    gcp_zone: str,
    *,
    beaker_cluster: str,
    beaker_priority: Priority,
    gcp_credentials_path: Optional[Path] = None,
) -> list[list[str]]:
    if beaker_cluster != "ai2/augusta":
        raise ValueError(
            "Only Augusta is supported. Making this work for other clusters probably would be a bad idea..."
        )

    if beaker_priority != Priority.urgent:
        log.warning(
            "Host selection depends on the cluster having nodes with jobs running at lower priority. "
            "It is relatively unlikely to work on non-urgent priorities."
        )

    assert num_nodes > 0
    assert num_nodes % num_execution_units == 0
    assert num_nodes % beaker_task_count == 0
    beaker_num_hosts_per_task = num_nodes // beaker_task_count

    hosts_metadata = get_hosts_metadata_from_gcp(gcp_zone, credentials_path=gcp_credentials_path)

    occupied_hosts = get_occupied_beaker_hosts(hosts_metadata, beaker_cluster, beaker_priority)

    return get_hostname_constraints(
        hosts_metadata,
        num_execution_units,
        beaker_num_hosts_per_task,
        beaker_task_count,
        occupied_hosts,
    )


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("num_nodes", type=int, help="Total number of nodes")
    parser.add_argument(
        "--num-execution-units",
        type=int,
        default=1,
        help="Number of `execution units`. An `execution unit` is abstraction for any node-using entity of which 1 or more copies are run that requires its nodes come from the same block (e.g., a model replica).",
    )
    parser.add_argument(
        "--task-count",
        type=int,
        default=1,
        help="Number of beaker (pre-replication) tasks this job is being spread across",
    )

    # Beaker-related settings
    parser.add_argument(
        "--priority",
        type=Priority,
        default=Priority.normal,
        help="Desired beaker job priority.",
    )
    parser.add_argument(
        "--cluster",
        type=str,
        default="ai2/augusta",
        help="The beaker cluster. This defaults to and is assumed to be Augusta for now.",
    )

    # GCP-related settings
    parser.add_argument(
        "--zone",
        type=str,
        default="us-central1-b",
        help="The GCP zone where the Augusta nodes are located.",
    )
    parser.add_argument(
        "--credentials-path",
        type=Path,
        help="The path to GCP credetials.",
    )
    args = parser.parse_args()

    print(
        get_beaker_hostname_constraints(
            args.num_nodes,
            args.num_execution_units,
            args.task_count,
            args.zone,
            beaker_cluster=args.cluster,
            beaker_priority=args.priority,
        )
    )


if __name__ == "__main__":
    main()



import logging
from dataclasses import dataclass
from functools import partial
from typing import Optional

import torch
import torch.nn as nn
from torch.distributed import DeviceMesh
from torch.distributed.tensor import Placement, Shard, distribute_module
from torch.distributed.tensor.parallel import SequenceParallel as _SequenceParallel

from olmo_core.config import Config

log = logging.getLogger(__name__)


@dataclass
class TensorParallelConfig(Config):
    """
    Configuration class for tensor parallelism (TP).
    """

    degree: int
    """
    The TP degree.
    """

    enable_async: bool = False
    """
    Enable experimental async tensor parallelism.
    """

    def maybe_enable_async_tp(self, tp_mesh: DeviceMesh):
        if self.enable_async:
            log.info("Enabling async tensor parallel")

            from torch.distributed._symmetric_memory import enable_symm_mem_for_group

            torch._inductor.config._micro_pipeline_tp = True  # type: ignore
            enable_symm_mem_for_group(tp_mesh.get_group().group_name)


class SequenceParallel(_SequenceParallel):
    def __init__(
        self,
        *,
        sequence_dim: int = 1,
        use_local_output: bool = False,
        output_layouts: Optional[Placement] = None,
    ):
        super().__init__(sequence_dim=sequence_dim, use_local_output=use_local_output)
        self.output_layouts = (output_layouts or Shard(sequence_dim),)

    @staticmethod
    def _prepare_output_fn(output_layouts, use_local_output, mod, outputs, device_mesh):
        del mod, device_mesh
        if outputs.placements != output_layouts:
            outputs = outputs.redistribute(placements=output_layouts, async_op=True)
        return outputs.to_local() if use_local_output else outputs

    def _apply(self, module: nn.Module, device_mesh: DeviceMesh) -> nn.Module:
        return distribute_module(
            module,
            device_mesh,
            self._replicate_module_fn,
            partial(self._prepare_input_fn, self.sequence_sharding),  # type: ignore
            partial(self._prepare_output_fn, self.output_layouts, self.use_local_output),  # type: ignore
        )



from .generation_module import GenerationModule
from .generation_module.config import GenerationConfig
from .generation_module.transformer import (
    TransformerGenerationModule,
    TransformerGenerationModuleConfig,
)

__all__ = [
    "GenerationConfig",
    "GenerationModule",
    "TransformerGenerationModule",
    "TransformerGenerationModuleConfig",
]



import datetime
import logging
import os
import random
import socket
import sys
from collections import deque
from typing import Any, Callable, Dict, Optional, Tuple

import torch.distributed as dist
import torch.multiprocessing as mp

from olmo_core.distributed.utils import (
    OLMO_LOCAL_RANK_ENV_VAR,
    OLMO_LOCAL_WORLD_SIZE_ENV_VAR,
    OLMO_NUM_NODES_ENV_VAR,
    init_distributed,
)

log = logging.getLogger(__name__)


_PORT_MIN = 29500
_PORT_MAX = 30000


def _initialize_ports() -> deque[int]:
    ports = list(range(_PORT_MIN, _PORT_MAX))
    random.Random().shuffle(ports)
    return deque(ports)


_PORTS = _initialize_ports()


def _port_in_use(host: str, port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.settimeout(1)
        return s.connect_ex((host, port)) == 0


def _get_next_port() -> int:
    global _PORTS
    port = _PORTS[0]
    _PORTS.rotate()
    return port


def _find_open_port(host: str = "127.0.0.1") -> int:
    port = _get_next_port()
    attempts = 0
    while _port_in_use(host, port):
        port += _get_next_port()
        attempts += 1
        if attempts >= 10:
            raise RuntimeError("failed to find an open port")
    return port


def _init_process(
    process_rank: int,
    world_size: int,
    backend: str,
    log_from_all_ranks: bool,
    func: Callable,
    func_args: Optional[Tuple[Any, ...]] = None,
    func_kwargs: Optional[Dict[str, Any]] = None,
    primary_addr: str = "127.0.0.1",
    primary_port: int = 29500,
):
    assert world_size > 1

    os.environ.setdefault(OLMO_NUM_NODES_ENV_VAR, "1")
    os.environ.setdefault(OLMO_LOCAL_WORLD_SIZE_ENV_VAR, str(world_size))
    os.environ.setdefault(OLMO_LOCAL_RANK_ENV_VAR, str(process_rank))

    init_distributed(
        backend=backend,
        timeout=datetime.timedelta(seconds=120),
        init_method=f"tcp://{primary_addr}:{primary_port}",
        world_size=world_size,
        rank=process_rank,
    )

    old_log_record_factory = logging.getLogRecordFactory()

    def log_record_factory(*args, **kwargs) -> logging.LogRecord:
        record = old_log_record_factory(*args, **kwargs)
        setattr(record, "local_rank", dist.get_rank())
        return record

    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(
        logging.Formatter(
            "[rank %(local_rank)s] %(asctime)s:%(name)s:%(lineno)s:%(levelname)s: %(message)s"
        )
    )
    logging.setLogRecordFactory(log_record_factory)

    if log_from_all_ranks or process_rank == 0:
        logging.basicConfig(level=logging.DEBUG, handlers=[handler])

    log.info("Starting test...")

    try:
        func(*(func_args or []), **(func_kwargs or {}))
    finally:
        dist.destroy_process_group()


def run_distributed_test(
    func: Callable,
    world_size: int = 2,
    log_from_all_ranks: bool = False,
    backend: str = "gloo",
    start_method: Optional[str] = None,
    func_args: Optional[Tuple[Any, ...]] = None,
    func_kwargs: Optional[Dict[str, Any]] = None,
    primary_addr: str = "127.0.0.1",
    primary_port: Optional[int] = None,
):
    """
    This runs the `func` in a simulated distributed environment.
    """
    if start_method is None:
        start_method = "fork" if backend == "gloo" else "spawn"

    if primary_port is None:
        primary_port = _find_open_port(host=primary_addr)

    log.info(f"Running distributed test on port {primary_port}...")

    mp.start_processes(
        _init_process,
        args=(
            world_size,
            backend,
            log_from_all_ranks,
            func,
            func_args,
            func_kwargs,
            primary_addr,
            primary_port,
        ),
        nprocs=world_size,
        start_method=start_method,
    )



import logging
from abc import abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, Optional, Tuple, Union

import torch
import torch.nn as nn
from torch.distributed import DeviceMesh
from torch.distributed.tensor import Placement, Replicate, Shard
from torch.distributed.tensor.parallel import (
    PrepareModuleInput,
    PrepareModuleOutput,
    parallelize_module,
)

from olmo_core.config import Config, DType, StrEnum
from olmo_core.distributed.parallel import (
    flatten_mesh,
    get_pp_stage_mesh,
    get_world_mesh,
)
from olmo_core.exceptions import OLMoConfigurationError
from olmo_core.ops import attach_auxiliary_loss

from ..buffer_cache import BufferCache
from ..feed_forward import FeedForwardConfig
from .loss import MoELoadBalancingLossGranularity
from .mlp import DroplessMoEMLP, MoEMLP
from .parallel_mlp import ParallelDroplessMLP, ParallelMLP, ParallelMLPBase
from .router import MoERouterConfig

if TYPE_CHECKING:
    from olmo_core.train.common import ReduceType

__all__ = ["MoEBase", "MoE", "DroplessMoE", "MoEConfig", "MoEType"]


log = logging.getLogger(__name__)


class MoEType(StrEnum):
    """
    An enumeration of the different MoE implementations.
    """

    default = "default"
    """
    ➡️ :class:`MoE`
    """

    dropless = "dropless"
    """
    ➡️ :class:`DroplessMoE`
    """


@dataclass
class MoEConfig(Config):
    name: MoEType = MoEType.default
    """
    The name of the implementation.
    """
    num_experts: int = 1
    hidden_size: int = 256
    capacity_factor: Optional[float] = None
    router: MoERouterConfig = field(default_factory=MoERouterConfig)
    shared_mlp: Optional[FeedForwardConfig] = None
    lb_loss_weight: Optional[float] = 0.01
    lb_loss_granularity: MoELoadBalancingLossGranularity = (
        MoELoadBalancingLossGranularity.local_batch
    )
    z_loss_weight: Optional[float] = None
    scale_loss_by_num_layers: bool = True
    dtype: DType = DType.float32

    def num_params(self, d_model: int) -> int:
        num_params = 0
        num_params += self.router.num_params(d_model, self.num_experts)
        num_params += 3 * d_model * self.hidden_size * self.num_experts
        if self.shared_mlp is not None:
            num_params += self.shared_mlp.num_params(d_model)
        return num_params

    def num_active_params(self, d_model: int) -> int:
        return (
            self.num_params(d_model)
            - (3 * d_model * self.hidden_size * self.num_experts)
            + (3 * d_model * self.hidden_size * self.router.top_k)
        )

    def build(
        self,
        d_model: int,
        *,
        n_layers: int = 1,
        init_device: str = "cpu",
        cache: Optional[BufferCache] = None,
    ) -> "MoEBase":
        kwargs = self.as_dict(exclude_none=True, recurse=False)
        kwargs.pop("name")
        kwargs.update(
            d_model=d_model,
            n_layers=n_layers,
            init_device=init_device,
            dtype=kwargs.pop("dtype").as_pt(),
            cache=cache,
        )

        try:
            if self.name == MoEType.default:
                return MoE(**kwargs)
            elif self.name == MoEType.dropless:
                return DroplessMoE(**kwargs)
            else:
                raise NotImplementedError(self.name)
        except TypeError as e:
            raise OLMoConfigurationError(
                f"invalid options for '{self.name}' {self.__class__.__name__}, {e}"
            ) from e


class MoEBase(nn.Module):
    """
    Base class for MoE implementations.
    """

    def __init__(
        self,
        *,
        d_model: int,
        num_experts: int,
        hidden_size: int,
        router: MoERouterConfig,
        shared_mlp: Optional[FeedForwardConfig] = None,
        init_device: str = "cpu",
        lb_loss_weight: Optional[float] = None,
        lb_loss_granularity: MoELoadBalancingLossGranularity = MoELoadBalancingLossGranularity.local_batch,
        z_loss_weight: Optional[float] = None,
        n_layers: int = 1,
        scale_loss_by_num_layers: bool = True,
        dtype: torch.dtype = torch.float32,
        cache: Optional[BufferCache] = None,
        **kwargs,
    ):
        super().__init__()
        if scale_loss_by_num_layers:
            if lb_loss_weight is not None:
                lb_loss_weight = lb_loss_weight / n_layers
            if z_loss_weight is not None:
                z_loss_weight = z_loss_weight / n_layers

        self.router = router.build(
            d_model,
            num_experts,
            lb_loss_weight=lb_loss_weight,
            lb_loss_granularity=lb_loss_granularity,
            z_loss_weight=z_loss_weight,
            dtype=dtype,
            init_device=init_device,
        )
        self.experts = self._init_parallel_mlp(
            d_model=d_model,
            num_experts=num_experts,
            hidden_size=hidden_size,
            dtype=dtype,
            init_device=init_device,
            cache=cache,
            **kwargs,
        )
        self.shared_mlp = (
            None
            if shared_mlp is None
            else shared_mlp.build(d_model, dtype=dtype, init_device=init_device)
        )
        self._ep_enabled = False

    @property
    def num_experts(self) -> int:
        return self.router.num_experts

    @property
    def top_k(self) -> int:
        return self.router.top_k

    @property
    def ep_enabled(self) -> bool:
        return self._ep_enabled

    def warmup_cache(self, max_local_microbatch_size: int):
        self.experts.warmup_cache(max_local_microbatch_size)

    def compute_metrics(
        self, reset: bool = True
    ) -> Dict[str, Tuple[torch.Tensor, Optional["ReduceType"]]]:
        return self.router.compute_metrics(reset=reset)

    def reset_metrics(self):
        self.router.reset_metrics()

    def post_batch(self, dry_run: bool = False):
        """
        Should be called right after the final backward of a complete batch but before the optimizer step.
        """
        self.router.post_batch(dry_run=dry_run)

    @abstractmethod
    def _init_parallel_mlp(
        self,
        *,
        d_model: int,
        num_experts: int,
        hidden_size: int,
        dtype: torch.dtype = torch.float32,
        init_device: str = "cpu",
        **kwargs,
    ) -> ParallelMLPBase:
        raise NotImplementedError

    def forward(
        self,
        x: torch.Tensor,
        *,
        loss_div_factor: Optional[Union[torch.Tensor, float]] = None,
    ) -> torch.Tensor:
        """
        Run the MoE on the input ``x`` of shape ``(*, d_model)``.

        :param x: The input of shape ``(*, d_model)``.

        :returns: The output of the MoE layer, the optional load-balancing loss, and the optional
            router Z-loss.
        """
        expert_weights, expert_indices, batch_size_per_expert, router_aux_loss = self.router(
            x, loss_div_factor=loss_div_factor
        )

        if router_aux_loss is not None:
            x = attach_auxiliary_loss(x, router_aux_loss)

        shared_out: Optional[torch.Tensor] = None
        if self.shared_mlp is not None:
            shared_out = self.shared_mlp(x)

        out = self.experts(x, expert_weights, expert_indices, batch_size_per_expert)

        if shared_out is not None:
            shared_out = shared_out / (self.top_k + 1)
            out = shared_out.add(out, alpha=self.top_k / (self.top_k + 1))

        return out

    def apply_pp(self, pp_mesh: DeviceMesh):
        world_mesh = get_world_mesh()
        assert world_mesh is not None
        stage_mesh = get_pp_stage_mesh(world_mesh, pp_mesh)
        group = flatten_mesh(stage_mesh).get_group()
        self.router.group = group

    def apply_ep(self, ep_mesh: DeviceMesh, **kwargs):
        """
        Apply expert parallelism.
        """
        self.experts.apply_ep(ep_mesh, **kwargs)
        self._ep_enabled = True

    def prepare_experts_for_fsdp(self, **kwargs):
        """
        Should be called before wrapping this module with FSDP2.
        """
        self.experts.prepare_experts_for_fsdp(**kwargs)

    def prepare_experts_for_ddp(self, **kwargs):
        """
        Should be called before wrapping this module with DDP2.
        """
        self.experts.prepare_experts_for_ddp(**kwargs)

    def apply_cp(self, cp_mesh: DeviceMesh):
        self.router.apply_cp(cp_mesh)

    def apply_tp(
        self,
        tp_mesh: DeviceMesh,
        input_layout: Optional[Placement] = None,
        output_layout: Optional[Placement] = None,
        use_local_output: bool = True,
        float8_enabled: bool = False,
    ):
        # Sequence parallel for the most part.
        parallelize_module(
            self,
            device_mesh=tp_mesh,
            parallelize_plan=PrepareModuleInput(
                input_layouts=None if input_layout is None else (input_layout,),
                desired_input_layouts=(Shard(1),),
                use_local_output=False,
            ),
        )

        # Sequence parallel.
        self.router.apply_tp(tp_mesh, float8_enabled=float8_enabled)

        # Expert parallel.
        self.experts.apply_tp(tp_mesh, float8_enabled=float8_enabled)

        # Model parallel.
        if self.shared_mlp is not None:
            self.shared_mlp.apply_tp(
                tp_mesh,
                input_layout=Shard(1),
                output_layout=Shard(1),
                use_local_output=True,
                float8_enabled=float8_enabled,
            )

        parallelize_module(
            self,
            device_mesh=tp_mesh,
            parallelize_plan=PrepareModuleOutput(
                output_layouts=(Shard(1),),
                desired_output_layouts=(output_layout or Replicate(),),
                use_local_output=use_local_output,
            ),
        )


class MoE(MoEBase):
    """
    A basic MoE implementation.
    """

    def __init__(
        self,
        *,
        d_model: int,
        num_experts: int,
        hidden_size: int,
        router: MoERouterConfig,
        shared_mlp: Optional[FeedForwardConfig] = None,
        capacity_factor: float = 1.2,
        init_device: str = "cpu",
        lb_loss_weight: Optional[float] = None,
        lb_loss_granularity: MoELoadBalancingLossGranularity = MoELoadBalancingLossGranularity.local_batch,
        z_loss_weight: Optional[float] = None,
        scale_loss_by_num_layers: bool = True,
        n_layers: int = 1,
        dtype: torch.dtype = torch.float32,
        cache: Optional[BufferCache] = None,
    ):
        super().__init__(
            d_model=d_model,
            num_experts=num_experts,
            hidden_size=hidden_size,
            router=router,
            shared_mlp=shared_mlp,
            init_device=init_device,
            lb_loss_weight=lb_loss_weight,
            lb_loss_granularity=lb_loss_granularity,
            z_loss_weight=z_loss_weight,
            scale_loss_by_num_layers=scale_loss_by_num_layers,
            n_layers=n_layers,
            dtype=dtype,
            capacity_factor=capacity_factor,
            cache=cache,
        )

    def _init_parallel_mlp(  # type: ignore[override]
        self,
        *,
        d_model: int,
        num_experts: int,
        hidden_size: int,
        capacity_factor: float,
        dtype: torch.dtype = torch.float32,
        init_device: str = "cpu",
        cache: Optional[BufferCache] = None,
    ) -> ParallelMLP:
        return ParallelMLP(
            mlp=MoEMLP(
                d_model=d_model,
                hidden_size=hidden_size,
                num_experts=num_experts,
                dtype=dtype,
                init_device=init_device,
            ),
            top_k=self.router.top_k,
            capacity_factor=capacity_factor,
            cache=cache,
        )


class DroplessMoE(MoEBase):
    """
    A dropless MoE implementation.
    """

    def _init_parallel_mlp(  # type: ignore[override]
        self,
        *,
        d_model: int,
        num_experts: int,
        hidden_size: int,
        dtype: torch.dtype = torch.float32,
        init_device: str = "cpu",
        cache: Optional[BufferCache] = None,
    ) -> ParallelDroplessMLP:
        return ParallelDroplessMLP(
            mlp=DroplessMoEMLP(
                d_model=d_model,
                num_experts=num_experts,
                hidden_size=hidden_size,
                dtype=dtype,
                init_device=init_device,
            ),
            top_k=self.router.top_k,
            cache=cache,
        )



from dataclasses import dataclass

from olmo_core.config import Config


@dataclass
class ContextParallelConfig(Config):
    """
    Configuration class for context parallelism (CP).
    """

    degree: int
    """
    The CP degree.
    """



import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Optional

import torch
import torch.distributed as dist
import torch.distributed.checkpoint.state_dict as dist_cp_sd

from olmo_core.aliases import PathOrStr
from olmo_core.config import Config, DType
from olmo_core.doc_utils import beta_feature
from olmo_core.float8 import Float8Config
from olmo_core.generate.generation_module.config import GenerationConfig
from olmo_core.nn.transformer import TransformerConfig
from olmo_core.train.train_module.transformer.config import (
    TransformerDataParallelConfig,
)
from olmo_core.utils import log_or_print

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from . import TransformerGenerationModule


@beta_feature
@dataclass
class TransformerGenerationModuleConfig(Config):
    """
    A configuration class for building :class:`TransformerGenerationModule` instances.
    """

    # Generation settings.
    generation_config: GenerationConfig
    # Model settings.
    compile_model: bool = False
    float8_config: Optional[Float8Config] = None
    dp_config: Optional[TransformerDataParallelConfig] = None

    # Checkpoint settings.
    state_dict_load_opts: Optional[Dict[str, Any]] = None
    load_key_mapping: Optional[Dict[str, str]] = None

    # Other settings.
    dtype: Optional[DType] = None
    """The dtype to build the model in."""

    def build(
        self,
        checkpoint_dir: PathOrStr,
        transformer_config: Optional[TransformerConfig] = None,
        device: Optional[torch.device] = None,
        process_group: Optional[dist.ProcessGroup] = None,
        work_dir: Optional[PathOrStr] = None,
        pre_download: bool = True,
        load_thread_count: Optional[int] = None,
    ) -> "TransformerGenerationModule":
        """
        Build the corresponding :class:`TransformerGenerationModule`.

        :param checkpoint_dir: Checkpoint directory to load from.
        :param transformer_config: The :class:`~olmo_core.nn.transformer.TransformerConfig` to use for generation.
        :param device: The device to use for generation.
        :param process_group: The process group for distributed loading.
        :param work_dir: Working directory for temporary files during loading.
        :param pre_download: Whether to pre-download remote checkpoints.
        :param load_thread_count: Number of threads to use for loading.
        """
        from olmo_core.generate.generation_module import TransformerGenerationModule

        config_dict = self.as_dict(exclude_none=True, recurse=False)
        if (dtype := config_dict.pop("dtype", None)) is not None:
            dtype = DType(dtype)
        if (autocast_precision := config_dict.pop("autocast_precision", None)) is not None:
            config_dict["autocast_precision"] = DType(autocast_precision).as_pt()
        if (state_dict_save_opts := config_dict.pop("state_dict_save_opts", None)) is not None:
            config_dict["state_dict_save_opts"] = dist_cp_sd.StateDictOptions(
                **state_dict_save_opts
            )
        if (state_dict_load_opts := config_dict.pop("state_dict_load_opts", None)) is not None:
            config_dict["state_dict_load_opts"] = dist_cp_sd.StateDictOptions(
                **state_dict_load_opts
            )
        log_or_print(log, f"TransformerGenerationModuleConfig: {config_dict}")

        return TransformerGenerationModule.from_checkpoint(
            checkpoint_dir=checkpoint_dir,
            transformer_config=transformer_config,
            device=device,
            process_group=process_group,
            work_dir=work_dir,
            pre_download=pre_download,
            load_thread_count=load_thread_count,
            dtype=dtype,
            **config_dict,
        )



"""
Utilities for training in Float8 via `torchao <https://github.com/pytorch/ao>`_.
"""

import logging
from dataclasses import dataclass
from typing import Optional, Set

import torch
import torch.nn as nn

from ..config import Config
from ..exceptions import OLMoConfigurationError
from .ao import AOFloat8LinearConfig, AOFloat8LinearRecipe

__all__ = ["Float8Config", "AOFloat8LinearConfig", "AOFloat8LinearRecipe"]

log = logging.getLogger(__name__)


@dataclass
class Float8Config(Config):
    """
    A configuration class for specifying Float8 options.

    :param ao: A torchao ``Float8Linear`` linear configuration.
    :param ao_recipe: Alternatively you can specify a recipe name from torchao.
    :param enabled: If ``False`` this will be a no-op.
    """

    ao: Optional[AOFloat8LinearConfig] = None
    ao_recipe: Optional[AOFloat8LinearRecipe] = None

    enabled: bool = True

    @property
    def should_precompute_float8_dynamic_scale_for_fsdp(self):
        if self.ao_recipe is not None:
            return False

        float8_linear_config = (
            self.ao if self.ao is not None else AOFloat8LinearConfig()
        ).to_ao_type()
        return float8_linear_config.enable_fsdp_float8_all_gather

    def apply_float8_linear(
        self, model: nn.Module, *, modules_to_ignore: Optional[Set[str]] = None
    ):
        """
        This method converts the linear layers of ``model`` to ``Float8Linear``.

        .. warning::
            This will mutate the model in place.

        .. warning::
            This should be called before compiling the model, applying activation checkpointing,
            or wrapping it with FSDP(2) or any other parallel wrapper.
        """
        if not self.enabled:
            return

        if not (torch.cuda.is_available() and torch.cuda.get_device_capability() >= (8, 9)):
            raise RuntimeError("Float8 training is only supported on SM89 or later")

        if self.ao_recipe is not None and self.ao is not None:
            raise OLMoConfigurationError("'ao_recipe' and 'ao' config are mutually exclusive")

        from torchao.float8 import Float8LinearConfig, convert_to_float8_training

        ignored_modules_found = set()

        def module_filter_fn(m: nn.Module, fqn: str) -> bool:
            nonlocal ignored_modules_found
            if modules_to_ignore is not None and fqn in modules_to_ignore:
                ignored_modules_found.add(fqn)
                return False

            # Linear layers must have all dimensions divisible by 16.
            if isinstance(m, nn.Linear):
                for d in m.weight.shape:
                    if d % 16 != 0:
                        return False

            return True

        # NOTE: there's a bug with `Float8Linear.from_float()` where it will override `requires_grad=False`
        # when `enable_fsdp_float8_all_gather=True`. So we have to reset frozen params after the fact.
        # https://github.com/pytorch/ao/issues/1871
        frozen_params: Set[str] = set()
        for n, p in model.named_parameters():
            if not p.requires_grad:
                frozen_params.add(n)

        # Mutates the model in place, replacing instances of nn.Linear with Float8Linear.
        float8_linear_config: Float8LinearConfig
        if self.ao_recipe is not None:
            float8_linear_config = Float8LinearConfig.from_recipe_name(self.ao_recipe.to_ao_type())
        else:
            float8_linear_config = (
                self.ao if self.ao is not None else AOFloat8LinearConfig()
            ).to_ao_type()

        convert_to_float8_training(
            model,
            config=float8_linear_config,
            module_filter_fn=module_filter_fn,
        )

        if modules_to_ignore is not None and modules_to_ignore != ignored_modules_found:
            raise OLMoConfigurationError(
                f"invalid module name(s) in 'modules_to_ignore': {list(modules_to_ignore - ignored_modules_found)}"
            )

        for n in frozen_params:
            p = model.get_parameter(n)
            p.requires_grad = False



from .config import GenerationConfig
from .generation_module import GenerationModule
from .transformer.config import TransformerGenerationModuleConfig
from .transformer.generation_module import TransformerGenerationModule

__all__ = [
    "GenerationConfig",
    "GenerationModule",
    "TransformerGenerationModule",
    "TransformerGenerationModuleConfig",
]



import torch


class AutoAuxiliaryLoss(torch.autograd.Function):
    """
    An autograd function that triggers the backward pass for an auxiliary loss.
    """

    @staticmethod
    def forward(ctx, activation: torch.Tensor, aux_loss: torch.Tensor):
        ctx.save_for_backward(aux_loss)
        return activation

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        (aux_loss,) = ctx.saved_tensors
        aux_loss_grad = torch.ones_like(aux_loss)
        return grad_output, aux_loss_grad


def attach_auxiliary_loss(activation: torch.Tensor, aux_loss: torch.Tensor) -> torch.Tensor:
    """
    Attach an auxiliary loss to an activation with an autograd function in order to trigger
    gradients for the aux loss in the backwards pass.

    :returns: The input activation unchanged.
    """
    return AutoAuxiliaryLoss.apply(activation, aux_loss)  # type: ignore[return-value]



"""
Configuration helpers for LLM cooking.

Indended to aid those migrating from olmo-cookbook to OLMo-core. Implements the same
defaults as olmo-cookbook where possible, and even includes some QOL improvements.
"""

import logging
from typing import Dict, Optional

import torch

from olmo_core.config import DType
from olmo_core.distributed.parallel import DataParallelType
from olmo_core.float8 import AOFloat8LinearConfig, Float8Config
from olmo_core.internal.common import get_beaker_username
from olmo_core.io import dir_is_empty
from olmo_core.optim import OptimGroupOverride, SkipStepAdamWConfig
from olmo_core.optim.scheduler import Scheduler
from olmo_core.train import Duration, LoadStrategy, TrainerConfig
from olmo_core.train.callbacks import (
    BeakerCallback,
    Callback,
    CheckpointerCallback,
    ConfigSaverCallback,
    GarbageCollectorCallback,
    GPUMemoryMonitorCallback,
    MonkeyPatcherCallback,
    ProfilerCallback,
    SlackNotifierCallback,
    WandBCallback,
)
from olmo_core.train.train_module import (
    TransformerActivationCheckpointingConfig,
    TransformerActivationCheckpointingMode,
    TransformerContextParallelConfig,
    TransformerDataParallelConfig,
    TransformerTrainModuleConfig,
)

log = logging.getLogger(__name__)


def configure_train_module(
    max_sequence_length: int,
    rank_microbatch_size: int,
    learning_rate: float,
    scheduler: Scheduler,
    float8_enabled: bool = False,
    activation_memory_budget: float = 1.0,  # smaller memory budget means more checkpointing
    dp_shard_degree: Optional[int] = None,
    cp_degree: Optional[int] = None,
) -> TransformerTrainModuleConfig:
    if not (0.0 < activation_memory_budget <= 1.0):
        raise ValueError("activation_memory_budget must be in the range [0.0, 1.0].")
    return TransformerTrainModuleConfig(
        rank_microbatch_size=rank_microbatch_size,
        max_sequence_length=max_sequence_length,
        optim=SkipStepAdamWConfig(
            lr=learning_rate,
            weight_decay=0.1,
            betas=(0.9, 0.95),
            group_overrides=[
                OptimGroupOverride(params=["embeddings.weight"], opts=dict(weight_decay=0.0))
            ],
            foreach=True,
        ),
        scheduler=scheduler,
        z_loss_multiplier=1e-5,
        max_grad_norm=1.0,
        compile_model=True,
        dp_config=TransformerDataParallelConfig(
            name=DataParallelType.hsdp,
            param_dtype=DType.bfloat16,
            reduce_dtype=DType.float32,
            shard_degree=dp_shard_degree,
        ),
        cp_config=(
            TransformerContextParallelConfig.llama3(degree=cp_degree, head_stride=4)
            if cp_degree
            else None
        ),
        ac_config=TransformerActivationCheckpointingConfig(
            mode=TransformerActivationCheckpointingMode.budget,
            activation_memory_budget=activation_memory_budget,
        )
        if activation_memory_budget < 1.0
        else None,
        float8_config=Float8Config(
            enabled=float8_enabled,
            ao=AOFloat8LinearConfig(
                enable_fsdp_float8_all_gather=True,
                force_recompute_fp8_weight_in_bwd=True,
                round_scales_to_power_of_2=True,
            ),
        ),
    )


def configure_trainer(
    max_duration: Duration,
    checkpoint_dir: str,
    work_dir: str,
    load_path: Optional[str] = None,
    load_trainer_state: Optional[bool] = None,
    load_optim_state: Optional[bool] = None,
    hard_stop: Optional[Duration] = None,
) -> TrainerConfig:
    load_strategy = LoadStrategy.always if load_path else LoadStrategy.if_available
    if load_path and dir_is_empty(load_path):
        raise FileNotFoundError(f"{load_path=} was provided, but the directory is empty.")
    trainer_config = TrainerConfig(
        max_duration=max_duration,
        load_path=load_path,
        save_folder=checkpoint_dir,
        work_dir=work_dir,
        hard_stop=hard_stop,
        load_strategy=load_strategy,
        load_trainer_state=load_trainer_state,
        load_optim_state=load_optim_state,
        save_overwrite=True,
    )
    return trainer_config


def configure_default_callbacks(
    run_name: str,
    wandb_group_name: str,
    wandb_project: str = "olmo-cookbook",
    checkpoint_save_interval: int = 1000,
) -> Dict[str, Callback]:
    callbacks = {
        "checkpointer": CheckpointerCallback(
            save_interval=checkpoint_save_interval,
            ephemeral_save_interval=100,
            save_async=False,  # TODO: enable async saving when augusta stops being silly
        ),
        "config_saver": ConfigSaverCallback(),
        "profiler": ProfilerCallback(enabled=False),
        "garbage_collector": GarbageCollectorCallback(),
        "slack_notifier": SlackNotifierCallback(name=run_name, enabled=False),
        "monkey_patcher": MonkeyPatcherCallback(),
        "wandb": WandBCallback(
            name=run_name,
            group=wandb_group_name,
            project=wandb_project,
            entity="ai2-llm",
            cancel_check_interval=20,
            enabled=True,
        ),
    }

    beaker_user = get_beaker_username()
    if beaker_user is not None:
        callbacks["beaker"] = BeakerCallback()
    if torch.cuda.is_available():
        callbacks["gpu_monitor"] = GPUMemoryMonitorCallback()

    return callbacks



import logging
import os
from dataclasses import dataclass
from typing import Optional

import requests

from olmo_core.aliases import PathOrStr
from olmo_core.config import StrEnum
from olmo_core.distributed.utils import get_rank
from olmo_core.exceptions import OLMoEnvironmentError

from .callback import Callback

log = logging.getLogger(__name__)

SLACK_WEBHOOK_URL_ENV_VAR = "SLACK_WEBHOOK_URL"
BEAKER_JOB_ID_ENV_VAR = "BEAKER_JOB_ID"
EXC_LINE_LIMIT = 30


class SlackNotificationSetting(StrEnum):
    """
    Defines the notifications settings for the Slack notifier callback.
    """

    all = "all"
    """
    Send all types notifications.
    """

    end_only = "end_only"
    """
    Only send a notification when the experiment ends (successfully or with a failure).
    """

    failure_only = "failure_only"
    """
    Only send a notification when the experiment fails.
    """

    none = "none"
    """
    Don't send any notifications.
    """


@dataclass
class SlackNotifierCallback(Callback):
    name: Optional[str] = None
    """
    A name to give the run.
    """

    notifications: SlackNotificationSetting = SlackNotificationSetting.end_only
    """
    The notification settings.
    """

    enabled: bool = True
    """
    Set to false to disable this callback.
    """

    webhook_url: Optional[str] = None
    """
    The webhook URL to post. If not set, will check the environment variable ``SLACK_WEBHOOK_URL``.
    """

    def post_attach(self):
        if not self.enabled or get_rank() != 0:
            return

        if self.webhook_url is None and SLACK_WEBHOOK_URL_ENV_VAR not in os.environ:
            raise OLMoEnvironmentError(f"missing env var '{SLACK_WEBHOOK_URL_ENV_VAR}'")

    def pre_train(self):
        if not self.enabled or get_rank() != 0:
            return

        if self.notifications == SlackNotificationSetting.all:
            self._post_message("started")

    def post_checkpoint_saved(self, path: PathOrStr):
        if not self.enabled or get_rank() != 0:
            return

        if self.notifications == SlackNotificationSetting.all:
            self._post_message(f"saved a checkpoint to {path}")

    def post_epoch(self):
        if not self.enabled or get_rank() != 0:
            return

        if self.notifications == SlackNotificationSetting.all:
            self._post_message("completed an epoch")

    def post_train(self):
        if not self.enabled or get_rank() != 0:
            return

        if self.notifications in (
            SlackNotificationSetting.all,
            SlackNotificationSetting.end_only,
        ):
            if self.trainer.is_canceled:
                self._post_message("canceled")
            else:
                self._post_message("completed successfully")

    def on_error(self, exc: BaseException):
        if not self.enabled or get_rank() != 0:
            return

        if self.notifications in (
            SlackNotificationSetting.all,
            SlackNotificationSetting.end_only,
            SlackNotificationSetting.failure_only,
        ):
            exc_lines = str(exc).rstrip("\n").split("\n")
            if len(exc_lines) > EXC_LINE_LIMIT:
                exc_lines = exc_lines[:EXC_LINE_LIMIT]
                exc_lines.append("...")
            exc_str = "\n".join(exc_lines)
            self._post_message(f"failed with error:\n```\n{exc_str}\n```")

    def _post_message(self, msg: str):
        webhook_url = self.webhook_url or os.environ.get(SLACK_WEBHOOK_URL_ENV_VAR)
        if webhook_url is None:
            raise OLMoEnvironmentError(f"missing env var '{SLACK_WEBHOOK_URL_ENV_VAR}'")

        progress = (
            f"*Progress:*\n"
            f"- step: {self.step:,d}\n"
            f"- epoch: {self.trainer.epoch}\n"
            f"- tokens: {self.trainer.global_train_tokens_seen:,d}"
        )

        if self.name is not None:
            msg = f"Run `{self.name}` {msg}\n{progress}"
        else:
            msg = f"Run {msg}\n{progress}"

        if BEAKER_JOB_ID_ENV_VAR in os.environ:
            msg = f"{msg}\n*Beaker job:* https://beaker.org/job/{os.environ[BEAKER_JOB_ID_ENV_VAR]}"

        try:
            requests.post(webhook_url, json={"text": msg})
        except Exception as e:
            log.exception(f"Failed to send Slack notification: {e}")



from dataclasses import dataclass
from functools import cached_property
from typing import Any, Callable, Iterable, List, Optional, Tuple

import torch
import torch.nn as nn
from torch.distributed import DeviceMesh
from torch.distributed.pipelining import PipelineStage
from torch.distributed.pipelining.schedules import (
    PipelineScheduleMulti,
    PipelineScheduleSingle,
    _PipelineSchedule,
    get_schedule_class,
)

from olmo_core.config import Config, StrEnum
from olmo_core.exceptions import OLMoConfigurationError


class PipelineSplitStyle(StrEnum):
    loop = "loop"
    v = "v"


class PipelineScheduleType(StrEnum):
    """
    An enumeration of the different pipeline schedules available.

    .. warning::
        The zero-bubble variants have several issues at the moment including not being compatible
        with ``torch.compile``.

    """

    # See torch.distributed.pipelining.schedules.get_schedule_class for a list of available.

    single_1F1B = "1F1B"
    interleaved_1F1B = "Interleaved1F1B"
    gpipe = "GPipe"
    looped_bfs = "LoopedBFS"
    interleaved_zero_bubble = "InterleavedZeroBubble"
    zbv_zero_bubble = "ZBVZeroBubble"

    @property
    def is_single_stage(self) -> bool:
        try:
            return issubclass(get_schedule_class(self), PipelineScheduleSingle)
        except ValueError as e:
            raise OLMoConfigurationError(f"Invalid pipeline schedule '{self}'") from e

    @property
    def is_multi_stage(self) -> bool:
        return not self.is_single_stage

    @property
    def default_style(self) -> PipelineSplitStyle:
        if self == self.zbv_zero_bubble:
            return PipelineSplitStyle.v
        else:
            return PipelineSplitStyle.loop


@dataclass
class PipelineParallelConfig(Config):
    """
    Configuration class for pipeline parallelism (PP).
    """

    degree: int
    """
    The PP degree.
    """

    schedule: PipelineScheduleType = PipelineScheduleType.interleaved_1F1B
    """
    The name of the schedule.
    """

    style: Optional[PipelineSplitStyle] = None
    """
    The split style.
    """

    def infer_style(self) -> PipelineSplitStyle:
        if self.style is not None:
            return self.style
        else:
            return self.schedule.default_style

    def final_stage_rank(self) -> int:
        style = self.infer_style()
        if style == PipelineSplitStyle.loop:
            return self.degree - 1
        elif style == PipelineSplitStyle.v:
            return 0
        else:
            raise NotImplementedError(style)

    def rank_completion_order(self) -> Iterable[int]:
        """
        The order that ranks within the PP group will complete a batch.
        """
        style = self.infer_style()
        if style == PipelineSplitStyle.loop:
            return range(self.degree - 1, -1, -1)
        elif style == PipelineSplitStyle.v:
            return range(self.degree)
        else:
            raise NotImplementedError(style)

    def stage_ids_this_rank(self, pp_rank: int, num_stages: int) -> Tuple[int, ...]:
        """
        Compute the stage ids for the stages that will run on this pp rank for either a looped or
        V style schedule.
        """
        style = self.infer_style()
        if num_stages % self.degree != 0:
            raise OLMoConfigurationError(
                f"num_stages {num_stages} must be evenly divisible by pipeline size {self.degree}"
            )

        stages_per_rank = num_stages // self.degree
        if style == PipelineSplitStyle.loop:
            return tuple(pp_rank + s * self.degree for s in range(stages_per_rank))
        elif style == PipelineSplitStyle.v:
            assert (
                stages_per_rank == 2
            ), f"v schedules assume 2 stages per rank, got {stages_per_rank}"
            stage_v_pairs = list(
                zip(range(self.degree), range(num_stages - 1, self.degree - 1, -1))
            )
            return stage_v_pairs[pp_rank]
        else:
            raise NotImplementedError(style)


class PipelineSchedule:
    """
    A thin wrapper around PyTorch pipeline schedule classes.

    :param n_microbatches: How many microbatches to split the global training batch into.
        If global training batch size must be evenly divisible by this.
        If not specified, the default will be the number of pipeline stages.
    """

    def __init__(
        self,
        *,
        model_parts: List[nn.Module],
        stages: List[PipelineStage],
        pp_mesh: DeviceMesh,
        schedule_name: PipelineScheduleType,
        loss_fn: Optional[Callable[[Any, torch.Tensor], torch.Tensor]] = None,
        num_microbatches: Optional[int] = None,
    ):
        self.model_parts = model_parts
        self.stages = stages
        self.pp_mesh = pp_mesh
        self.loss_fn = loss_fn

        try:
            schedule_class = get_schedule_class(schedule_name)
        except ValueError as e:
            raise OLMoConfigurationError(f"Invalid pipeline schedule name '{schedule_name}'") from e

        if num_microbatches is None:
            num_microbatches = pp_mesh.size()

        schedule: _PipelineSchedule
        if issubclass(schedule_class, PipelineScheduleSingle):
            if len(model_parts) > 1:
                raise OLMoConfigurationError(
                    f"Expected a single stage for '{schedule_name}' pipeline schedule"
                )
            schedule = schedule_class(
                stages[0], n_microbatches=num_microbatches, loss_fn=self.loss_fn
            )
        elif issubclass(schedule_class, PipelineScheduleMulti):
            schedule = schedule_class(
                stages,  # type: ignore[arg-type]
                n_microbatches=num_microbatches,
                loss_fn=self.loss_fn,
            )
        else:
            raise NotImplementedError(schedule_class)

        self.base_schedule = schedule
        self.num_microbatches = num_microbatches

    @cached_property
    def has_first_stage(self) -> bool:
        for stage in self.stages:
            if stage.is_first:
                return True
        return False

    @cached_property
    def has_last_stage(self) -> bool:
        for stage in self.stages:
            if stage.is_last:
                return True
        return False

    def step(
        self,
        *args,
        target: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> Tuple[Any, Optional[torch.Tensor]]:
        """
        :param args: Only passed to first stage.
        :param kwargs: Passed to all stages.
        """
        losses: Optional[List[torch.Tensor]] = None
        if self.has_last_stage and self.loss_fn is not None:
            losses = []
        else:
            target = None

        output = self.base_schedule.step(*args, target=target, losses=losses, **kwargs)
        return output, None if losses is None else torch.stack(losses)



from typing import Type, Union

import numpy as np

from olmo_core.config import StrEnum

NumpyUIntTypes = Union[Type[np.uint8], Type[np.uint16], Type[np.uint32], Type[np.uint64]]


class LongDocStrategy(StrEnum):
    """
    Specifies how to handle documents that are longer than the max sequence length when packing.
    """

    truncate = "truncate"
    """
    Long docs are truncated and the excess tokens are discarded.
    """

    fragment = "fragment"
    """
    Long docs are split into smaller docs so that no tokens are discarded, but you end up with
    fragmented docs.
    """


class NumpyDatasetDType(StrEnum):
    """
    Supported numpy unsigned integer data types for datasets.
    """

    uint8 = "uint8"
    uint16 = "uint16"
    uint32 = "uint32"
    uint64 = "uint64"

    def as_np_dtype(self) -> NumpyUIntTypes:
        """
        Convert the enum value to its corresponding numpy dtype.

        Returns:
            The numpy unsigned integer dtype corresponding to this enum value.
        """
        return getattr(np, str(self))



import math
from dataclasses import dataclass
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributed import DeviceMesh
from torch.distributed.tensor.parallel import parallelize_module
from torch.distributed.tensor.placement_types import Placement, Replicate

from ..config import Config, DType, StrEnum
from ..doc_utils import beta_feature
from ..exceptions import OLMoConfigurationError
from .functional import l2_normalize
from .utils import get_tp_wrappers

__all__ = ["FeedForwardType", "FeedForwardConfig", "FeedForward", "NormalizedFeedForward"]


class FeedForwardType(StrEnum):
    """
    An enumeration of the different feed-forward / MLP implementations.
    """

    default = "default"
    """
    ➡️ :class:`FeedForward`
    """

    normalized = "normalized"
    """
    ➡️ :class:`NormalizedFeedForward`
    """


@dataclass
class FeedForwardConfig(Config):
    """
    A config for building :class:`FeedForward` modules.
    """

    hidden_size: int
    name: FeedForwardType = FeedForwardType.default
    """
    The name of the implementation.
    """
    bias: Optional[bool] = None
    dtype: Optional[DType] = None

    def num_params(self, d_model: int) -> int:
        """
        The number of params that the module will have once built.

        :param d_model: The model dimensionality.
        """
        bias = self.bias if self.bias is not None else self.name != FeedForwardType.normalized

        params = 0

        params += 3 * d_model * self.hidden_size
        if bias:
            params += 2 * self.hidden_size + d_model

        # w1 + w3 scaling factors
        if self.name == FeedForwardType.normalized:
            params += 2 * self.hidden_size

        return params

    def build(
        self, d_model: int, *, dtype: Optional[torch.dtype] = None, init_device: str = "cpu"
    ) -> "FeedForward":
        """
        Build the corresponding feed-forward module.

        :param d_model: The model dimensionality.
        :param init_device: The device initialize the parameters on, e.g. "cpu", "meta".
        """
        kwargs = self.as_dict(exclude_none=True)
        kwargs.pop("name")
        kwargs.update(d_model=d_model, init_device=init_device)
        if self.dtype is not None:
            kwargs["dtype"] = self.dtype.as_pt()
        elif dtype is not None:
            kwargs["dtype"] = dtype

        try:
            if self.name == FeedForwardType.default:
                return FeedForward(**kwargs)
            elif self.name == FeedForwardType.normalized:
                return NormalizedFeedForward(**kwargs)
            else:
                raise NotImplementedError(self.name)
        except TypeError as e:
            raise OLMoConfigurationError(
                f"invalid options for '{self.name}' {self.__class__.__name__}, {e}"
            ) from e


class FeedForward(nn.Module):
    """
    Basic feed-forward module with SwiGLU activation.
    """

    def __init__(
        self,
        *,
        d_model: int,
        hidden_size: int,
        bias: bool = True,
        dtype: torch.dtype = torch.float32,
        init_device: str = "cpu",
    ):
        super().__init__()
        self.d_model = d_model
        self.hidden_size = hidden_size
        self.w1 = nn.Linear(d_model, hidden_size, bias=bias, dtype=dtype, device=init_device)
        self.w2 = nn.Linear(hidden_size, d_model, bias=bias, dtype=dtype, device=init_device)
        self.w3 = nn.Linear(d_model, hidden_size, bias=bias, dtype=dtype, device=init_device)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Run the feed-forward on the input ``x``.

        :param x: The input of shape ``(*, d_model)``.
        """
        return self.w2(F.silu(self.w1(x)) * self.w3(x))

    def apply_tp(
        self,
        tp_mesh: DeviceMesh,
        input_layout: Optional[Placement] = None,
        output_layout: Optional[Placement] = None,
        use_local_output: bool = True,
        float8_enabled: bool = False,
    ):
        rowwise_parallel, colwise_parallel, prepare_module_input = get_tp_wrappers(
            float8_enabled=float8_enabled
        )

        parallelize_module(
            module=self,
            device_mesh=tp_mesh,
            parallelize_plan=prepare_module_input(
                input_layouts=None if input_layout is None else (input_layout,),
                desired_input_layouts=(Replicate(),),
            ),
        )

        parallelize_module(
            module=self,
            device_mesh=tp_mesh,
            parallelize_plan={
                "w1": colwise_parallel(),
                "w2": rowwise_parallel(
                    output_layouts=output_layout, use_local_output=use_local_output
                ),
                "w3": colwise_parallel(),
            },
        )


@beta_feature
class NormalizedFeedForward(FeedForward):
    """
    An nGPT feed-forward implementation.
    """

    def __init__(
        self,
        *,
        d_model: int,
        hidden_size: int,
        dtype: torch.dtype = torch.float32,
        init_device: str = "cpu",
    ):
        super().__init__(
            d_model=d_model,
            hidden_size=hidden_size,
            dtype=dtype,
            init_device=init_device,
            bias=False,
        )
        self.sw_init_value = 1.0
        self.sw_init_scaling = 1.0
        self.sw1 = torch.nn.Parameter(torch.empty(hidden_size, dtype=dtype, device=init_device))
        self.sw3 = torch.nn.Parameter(torch.empty(hidden_size, dtype=dtype, device=init_device))
        self.sqrt_d_model = math.sqrt(d_model)
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.ones_(self.sw1)
        nn.init.ones_(self.sw3)
        with torch.no_grad():
            self.sw1.mul_(self.sw_init_scaling)
            self.sw3.mul_(self.sw_init_scaling)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        sw1 = self.sw1 * ((self.sw_init_value / self.sw_init_scaling) * self.sqrt_d_model)
        sw3 = self.sw3 * (self.sw_init_value / self.sw_init_scaling)
        return self.w2(F.silu(sw1 * self.w1(x)) * (sw3 * self.w3(x)))

    def apply_tp(
        self,
        tp_mesh: DeviceMesh,
        input_layout: Optional[Placement] = None,
        output_layout: Optional[Placement] = None,
        use_local_output: bool = True,
        float8_enabled: bool = False,
    ):
        del tp_mesh, input_layout, output_layout, use_local_output, float8_enabled

        raise NotImplementedError(
            "TP is not implemented yet for the normalized feed-forward variant"
        )

    @torch.no_grad()
    def normalize_matrices(self):
        """
        Normalize the weights in all matrices. This should be called after each optimizer step, which
        the :class:`~olmo_core.train.train_module.TransformerTrainModule` will handle for you.
        """
        self._normalize_matrix(self.w1.weight)
        self._normalize_matrix(self.w2.weight, dim=0)
        self._normalize_matrix(self.w3.weight)

    def _normalize_matrix(self, w: torch.Tensor, dim: int = -1):
        w.copy_(l2_normalize(w, dim=dim))
